"""
    Interface to reader a jpeg using PyTurboJPEG module.

    Please note: you need to install the turbojpeg library on your system
    for the module to work, you can follow the installation instruction in https://github.com/lilohuang/PyTurboJPEG

"""

from qimview.utils.viewer_image import *
from qimview.utils.utils import get_time

try:
    from turbojpeg import TurboJPEG, TJPF_RGB, TJPF_BGR, TJFLAG_FASTDCT
except:
    has_turbojpeg = False
else:
    has_turbojpeg = True
import configparser
import os

config = configparser.ConfigParser()
config_file = os.path.expanduser('~/.qimview.cfg')
lib_path = None
if os.path.isfile(config_file): 
    config.read([config_file])
    try:
        lib_path = config['READER.TURBOJPEG']['LibPath']
    except Exception as e:
        pass

# Initialize once the TurboJPEG instance, as a global variable
try:
    if lib_path:
        gb_turbo_jpeg = TurboJPEG(lib_path)
    else:
        gb_turbo_jpeg = TurboJPEG()
except Exception as e:
    print(f"Failed to load TurboJPEG library")
    gb_turbo_jpeg = None

def read_jpeg_turbojpeg(image_filename, image_buffer, read_size='full', use_RGB=True, verbose=False):
    try:
        if verbose:
            start = get_time()
        # using default library installation
        # decoding input.jpg to BGR array
        downscale = {'full': 1, '1/2': 2, '1/4': 4, '1/8': 8}[read_size]
        scale = (1, downscale)

        # do we need the fast DCT?
        flags = TJFLAG_FASTDCT
        pixel_format = TJPF_RGB if use_RGB else TJPF_BGR

        if image_buffer is None:
            with open(image_filename, 'rb') as d:
                image_buffer = d.read()
        # if verbose:
        #     im_header = gb_turbo_jpeg.decode_header(image_buffer)
        #     print(f" header {im_header} {int(get_time() - start1) * 1000} ms")
        im = gb_turbo_jpeg.decode(image_buffer, pixel_format=pixel_format, scaling_factor=scale, flags=flags)

        if verbose:
            print(f" turbojpeg read ...{image_filename[-15:]} took {get_time() - start:0.3f} sec.")
        viewer_image = ViewerImage(im, precision=8, downscale=downscale, channels=ImageFormat.CH_RGB if use_RGB else ImageFormat.CH_BGR)
        return viewer_image
    except Exception as e:
        print("read_jpeg: Failed to load image with turbojpeg {0}: {1}".format(image_filename, e))
        return None




#
#
# started from https://cyrille.rossant.net/2d-graphics-rendering-tutorial-with-pyopengl/
#
# check also https://doc.qt.io/archives/4.6/opengl-overpainting.html
#

from ..utils.qt_imports    import QtWidgets
from .image_viewer         import trace_method
from .gl_image_viewer_base import GLImageViewerBase

import OpenGL
OpenGL.ERROR_ON_COPY = True
import OpenGL.GL as gl

import argparse
import sys


class GLImageViewer(GLImageViewerBase):

    def __init__(self, parent=None, event_recorder=None):
        self.event_recorder = event_recorder
        super().__init__(parent)
        self.setAutoFillBackground(False)
        self.textureID  = None
        self.tex_width, self.tex_height = 0, 0
        self.opengl_debug = True
        self.trace_calls  = False

    def initializeGL(self):
        """Initialize OpenGL, VBOs, upload data on the GPU, etc.
        """
        # self.setTexture()
        pass

    def viewer_update(self):
        self.update()

    def paintGL(self):
        self.paintAll()

    def myPaintGL(self):
        """Paint the scene.
        """
        if self.trace_calls:
            t = trace_method(self.tab)
        self.start_timing()
        if self.textureID is None:
            print("GLImageViewer paintGL not textureID")
            return
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)
        gl.glTexEnvi(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_DECAL)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.textureID)
        # gl.glGenerateMipmap (gl.GL_TEXTURE_2D)
        gl.glEnable(gl.GL_TEXTURE_2D)
        gl.glBegin(gl.GL_QUADS)

        x0, x1, y0, y1 = self.image_centered_position()
        x0 = int(x0)
        x1 = int(x1)
        y0 = int(y0)
        y1 = int(y1)
        # print("{} {} {} {}".format(x0,x1,y0,y1))

        gl.glTexCoord2i(0, 0)
        gl.glVertex2i(x0, y1)

        gl.glTexCoord2i(0, 1)
        gl.glVertex2i(x0, y0)

        gl.glTexCoord2i(1, 1)
        gl.glVertex2i(x1, y0)

        gl.glTexCoord2i(1, 0)
        gl.glVertex2i(x1, y1)

        gl.glEnd()

        gl.glDisable(gl.GL_TEXTURE_2D)
        gl.glTexEnvi(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_MODULATE)

        self.print_timing(add_total=True)
        self.opengl_error()

    def resizeGL(self, width, height):
        """Called upon window resizing: reinitialize the viewport.
        """
        # print("ResizeGL")
        if self.trace_calls:
            t = trace_method(self.tab)
        # size give for opengl are in pixels, qt uses device independent size otherwise
        print(f"self.devicePixelRatio() {self.devicePixelRatio()}")
        self._width = width*self.devicePixelRatio()
        self._height = height*self.devicePixelRatio()
        # print("width height ratios {} {}".format(self._width/self.width(), self._height/self.height()))
        self.viewer_update()

    def event(self, evt):
        if self.event_recorder is not None:
            self.event_recorder.store_event(self, evt)
        return super().event(evt)

if __name__ == '__main__':
    from qimview.image_readers import gb_image_reader

    # import numpy for generating random data points
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('input_image', help='input image')
    args = parser.parse_args()
    _params = vars(args)

    # define a Qt window with an OpenGL widget inside it
    # class TestWindow(QtGui.QMainWindow):
    class TestWindow(QtWidgets.QMainWindow):
        def __init__(self):
            super(TestWindow, self).__init__()
            self.widget = GLImageViewer(self)
            self.show()
        def load(self):
            im = gb_image_reader.read(_params['input_image'])
            self.widget.set_image(im)
            self.widget.image_name = _params['input_image']
            # put the window at the screen position (100, 100)
            self.setGeometry(0, 0, self.widget._width, self.widget._height)
            self.setCentralWidget(self.widget)

    # create the Qt App and window
    app = QtWidgets.QApplication(sys.argv)
    window = TestWindow()
    window.load()
    window.show()
    app.exec()



import pytest

@pytest.fixture
def input_value():
   input = 39
   return input

def test_divisible_by_3(input_value):
   assert input_value % 3 == 0

@pytest.mark.skip
def test_divisible_by_6(input_value):
   assert input_value % 6 == 0



""" 
    Deal with MultiView mouse events
"""

from typing import NewType, TypeVar, TYPE_CHECKING
from qimview.utils.qt_imports import QtGui, QtCore
if TYPE_CHECKING:
    from qimview.image_viewers.multi_view import MultiView
from .mouse_events import MouseEvents, MouseMotionActions


# T is a type that inherits from QWidget

class MultiViewMouseEvents(MouseEvents['MultiView']):
    """ Implement mouse events for MultiView """

    def __init__(self, multiview: 'MultiView'):
        super().__init__(multiview)

        self._mouse_callback.update({
            'Left Pressed'  : self.action_activate,
            'Left+DblClick' : self.toggle_show_single_viewer,
        })


    def action_activate(self,  event: QtGui.QMouseEvent) -> bool:
        """ Set viewer active """
        for v in self._widget.image_viewers:
            if v.geometry().contains(event.pos()):
                # Set the current viewer active before processing the double click event
                self._widget.on_active(v)
                self._widget.update_image()
                return True
        return False

    def toggle_show_single_viewer(self, event: QtGui.QMouseEvent)->bool:
        """ Show only the selected viewer or all viewers """
        # Need to find the viewer that has been double clicked
        for v in self._widget.image_viewers:
            if v.geometry().contains(event.pos()):
                # Set the current viewer active before processing the double click event
                self._widget.on_active(v)
                self._widget._show_active_only = not self._widget._show_active_only
                if not self._widget._active_viewer:
                    return False
                # Set the current image
                self._widget.output_label_reference_image = v.image_name
                # Update the image to show/hide viewers
                self._widget.update_image()
                return True
        return False

    def mouse_move_unpressed(self, event: QtGui.QMoveEvent)->None:
        """ Actions while moving the mouse without pressing any button """
        event.ignore()



import glob
import os
import argparse

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('dir', type=str, help='directory to look for suffixes')


    verbose = False
    args = parser.parse_args()
    _params = vars(args)

    im_list = glob.glob(f"{_params['dir']}/**/*.jpg", recursive=True)
    # print(im_list)
    bn_list = [ os.path.splitext(os.path.basename(n))[0] for n in im_list]

    def list_suffixes(name):
        res = []
        splitted = name.split('_')
        for n in range(1,len(splitted)):
            res.append('_'.join(splitted[n:]))
        return res

    suffixes = []
    for bn in bn_list:
        suffixes.extend(list_suffixes(bn))
    suffixes = list(set(suffixes))

    im_list_scores = {}
    for suf in suffixes:
        if verbose: print(f" ----- suffix {suf}")
        im_list =  list(filter(lambda s: s.endswith(suf), bn_list))
        if verbose: print(f"im_list {im_list}")
        new_bn_list = [ f[:-len(suf)] for f in im_list]
        list_str = ','.join(new_bn_list)
        # print(f"list_str {list_str}")
        if list_str not in im_list_scores:
            im_list_scores[list_str] = [suf]
        else:
            im_list_scores[list_str].append(suf)

    for k in im_list_scores:
        if verbose: print(f"{k}:  {im_list_scores[k]}")
    candidates = sorted(im_list_scores, key=lambda x: x.count(',')*len(im_list_scores[x]), reverse=True)
    for n in range(len(candidates)):
        if n<30:
            print("----------")
            c = candidates[n]
            print(','.join(im_list_scores[c]))
            print(f"{c.count(',')*len(im_list_scores[c])}")





import sys
import time
import numpy as np
from collections import deque
import os


def get_time():
	is_windows = sys.platform.startswith('win')
	if is_windows:
		if hasattr(time, 'clock'):
			return time.clock()
		else:
			return time.perf_counter()
	else:
		return time.time()

def clip_value(value, lower, upper):
    return lower if value < lower else upper if value > upper else value

def get_size(obj, seen=None):
	"""
	Recursively finds size of objects
	"""
	size = sys.getsizeof(obj)
	if seen is None:
		seen = set()
	obj_id = id(obj)
	if obj_id in seen:
		return 0
	# Important mark as seen *before* entering recursion to gracefully handle
	# self-referential objects
	seen.add(obj_id)
	if isinstance(obj, dict):
		size += sum([get_size(v, seen) for v in obj.values()])
		size += sum([get_size(k, seen) for k in obj.keys()])
	elif hasattr(obj, '__dict__'):
		size += get_size(obj.__dict__, seen)
	elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
		size += sum([get_size(i, seen) for i in obj])
	return size


# function found at https://code.tutsplus.com/tutorials/understand-how-much-memory-your-python-objects-use--cms-25609
def deep_getsizeof(o, ids) -> int:
	""" Find the memory footprint of a Python object
	This is a recursive function that drills down a Python object graph
	like a dictionary holding nested dictionaries with lists of lists
	and tuples and sets.

	The sys.getsizeof function does a shallow size of only. It counts each
	object inside a container as pointer only regardless of how big it
	really is.

	:param o: the object
	:param ids:
	:return:
	"""
	d = deep_getsizeof

	if id(o) in ids:
		return 0

	# print(f' deep_getsizeof() type {type(o)} getsizeof {sys.getsizeof(o)}')
	try:
		image = o.image
		if isinstance(image, np.ndarray):
			r = image.nbytes
	except Exception as e:
		if isinstance(o, np.ndarray):
			r = o.nbytes
			# print(f' nbytes {o.nbytes}')
		else:
			r = sys.getsizeof(o)

	ids.add(id(o))
	if isinstance(o, str) or isinstance(0, str):
		return r
	if isinstance(o, dict):
		return r + sum(d(k, ids) + d(v, ids) for k, v in o.items())

	if isinstance(o,  (tuple, list, set, deque)):
		# print "*"
		return r + sum(d(x, ids) for x in o)
	return r



""" 
    Deal with ImageViewer mouse events
"""

from typing import TypeVar, TYPE_CHECKING
from qimview.utils.qt_imports import QtGui, QtCore, QtWidgets
if TYPE_CHECKING:
    from qimview.image_viewers.image_viewer import ImageViewer
from .mouse_events import MouseEvents, MouseMotionActions

# T is a type that inherits from QWidget
V = TypeVar('V', bound='ImageViewer')

class MousePanActions(MouseMotionActions[V]):
    """ Panning while mouse button is pressed """
    def press(self, event : QtGui.QMouseEvent) -> None:
        """ Press event """
        super().press(event)
        event.setAccepted(True)
    def move(self, event: QtGui.QMoveEvent) -> None:
        """ Move event """
        super().move(event)
        self._widget.mouse_pos = event.pos()
        self._widget.mouse_displ = self._delta
        self._widget.viewer_update()
        self._widget.synchronize()
        event.accept()
    def release(self, event: QtGui.QMouseEvent) -> None:
        """ Release event """
        self._widget.current_dx = int(self._widget.check_translation()[0])
        self._widget.current_dy = int(self._widget.check_translation()[1])
        super().release(event)
        self._widget.mouse_displ = self._delta

class MouseZoomActions(MouseMotionActions[V]):
    """ Zooming while mouse button is pressed """
    def press(self, event : QtGui.QMouseEvent) -> None:
        """ Press event """
        super().press(event)
        event.setAccepted(True)
    def move(self, event: QtGui.QMoveEvent) -> None:
        """ Move event """
        super().move(event)
        self._widget.mouse_pos = event.pos()
        self._widget.mouse_zoom_displ = self._delta
        self._widget.viewer_update()
        self._widget.synchronize()
        event.accept()
    def release(self, event: QtGui.QMouseEvent) -> None:
        """ Release event """
        im = self._widget.get_image()
        if im is not None:
            self._widget.current_scale = self._widget.new_scale(-self._delta.y(),im.data.shape[0])
        super().release(event)
        self._widget.mouse_zoom_displ = self._delta

class ImageViewerMouseEvents(MouseEvents[V]):
    """ Implement events for ImageViewer """
 
    def __init__(self, viewer: V):
        super().__init__(viewer)

        # # Accept touch events
        # viewer._widget.setAttribute(QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents)

        self._mouse_callback.update({
            'Left+DblClick on histogram' : self.toogle_histo_size,
            # Touch pad panning will be interpreted as Wheel event
            'Wheel'                      : self.wheel_pan,
            'Shft+Wheel'                 : self.wheel_pan_horizontal,
            # Set Ctrl+Wheel for touch pad zooming with 2 fingers
            'Ctrl+Wheel'                 : self.wheel_zoom,
            'Move on text'               : self.tooltip_image_filename,
        })

        self._motion_classes.update({
            'Left Motion'      : MousePanActions,
            'Ctrl+Left Motion' : MouseZoomActions,
            'Middle Motion'    : MouseZoomActions,
        })

        def in_histo(point:QtCore.QPoint)->bool:
            if self._widget._histo_rect:
                return self._widget._histo_rect.contains(point)
            return False

        def in_text(point:QtCore.QPoint)->bool:
            if self._widget._text_rect:
                return self._widget._text_rect.contains(point)
            return False

        self._regions.update({
            'histogram'  : in_histo,
            'text'       : in_text,
        })

    def mouse_move_unpressed(self, event: QtGui.QMoveEvent)->None:
        """ Actions while moving the mouse without pressing any button """
        if self._widget.show_overlay:
            self._widget.mouse_pos = event.pos()
            self._widget.viewer_update()
            event.accept()
        elif self._widget.show_cursor:
            self._widget.mouse_pos = event.pos()
            self._widget.viewer_update()
            self._widget.synchronize()
            event.accept()
        else:
            event.ignore()

    def toogle_histo_size(self, _)->bool:
        """ Switch histogram scale from 1 to 3 """
        self._widget._histo_scale = (self._widget._histo_scale % 3) + 1 
        self._widget.viewer_update()
        return True

    def wheel_zoom(self, event: QtGui.QWheelEvent) -> bool:
        """ Zoom in/out based on wheel angle """
        # Zoom by applying a factor to the distances to the sides
        # print(f"{event.angleDelta(), event.deviceType(), event.source()}")
        delta = event.angleDelta().y()
        # print("delta = {}".format(delta))
        coeff = delta/5
        # coeff = 20 if delta > 0 else -20
        if im := self._widget.get_image():
            self._widget.current_scale = self._widget.new_scale(coeff, im.data.shape[0])
            self._widget.viewer_update()
            self._widget.synchronize()
            return True
        return False

    def wheel_pan(self, event: QtGui.QWheelEvent) -> bool:
        """ Pan (image translation) based on wheel angle, mainly for touchpad """
        # Zoom by applying a factor to the distances to the sides
        # print(f"{event.angleDelta(), event.deviceType(), event.source()}")
        delta = event.angleDelta()
        if im := self._widget.get_image():
            self._widget.mouse_displ = delta
            self._widget.viewer_update()
            self._widget.synchronize()
            # Update current displacement
            self._widget.current_dx = int(self._widget.check_translation()[0])
            self._widget.current_dy = int(self._widget.check_translation()[1])
            return True
        return False

    def wheel_pan_horizontal(self, event: QtGui.QWheelEvent) -> bool:
        """ Pan horizontal (image horizontal translation) based on wheel angle """
        # Zoom by applying a factor to the distances to the sides
        # print(f"{event.angleDelta(), event.deviceType(), event.source()}")
        delta = event.angleDelta()
        # To translation horizontally using wheel mouse
        delta.setX(delta.y())
        delta.setY(0)
        if im := self._widget.get_image():
            self._widget.mouse_displ = delta
            self._widget.viewer_update()
            self._widget.synchronize()
            # Update current displacement
            self._widget.current_dx = int(self._widget.check_translation()[0])
            self._widget.current_dy = int(self._widget.check_translation()[1])
            return True
        return False

    def tooltip_image_filename(self, event: QtGui.QMoveEvent) -> bool:
        """ Display image filename as tooltip """
        global_pos = self._widget.mapToGlobal(event.pos())
        QtWidgets.QToolTip.showText(global_pos, f"{self._widget._image.filename}", 
                                    self._widget, self._widget._text_rect)
        # Return False to allow event propagation
        return False



""" 
    Base class for mouse events
"""

from abc import abstractmethod
from typing import Optional, Dict, Callable, Generic, TypeVar, Type
from qimview.utils.qt_imports import QtGui, QtCore, QtWidgets

def add2repr(res:str, elt:str) -> str:
    """ Add a substring to a string representing an event """
    if res=='':
        return elt
    return res+'+'+elt

# T is a type that inherits from QWidget
T = TypeVar('T')

class MouseMotionActions(Generic[T]):
    """ Base class to deal with Mouse Motion with button (Press + Move + Release) events """
    def __init__(self, widget: T) -> None:
        self._press_pos : Optional[QtCore.QPoint] = None
        self._delta     : QtCore.QPoint = QtCore.QPoint(0,0)
        self._widget    : T = widget

    def press(self, event : QtGui.QMouseEvent) -> None:
        """ Press event """
        self._press_pos = event.pos()

    def move(self, event: QtGui.QMoveEvent) -> None:
        """ Move event """
        if self._press_pos:
            self._delta = event.pos() - self._press_pos

    def release(self, event: QtGui.QMouseEvent) -> None:
        """ Release event """
        self._delta = QtCore.QPoint(0,0)

class MouseEvents(Generic[T]):
    """ Implement events for ImageViewer

        Create a string representation of each event to associate a corresponding callback
        Modifier + Key + Mouse Event + Position
         - modifier:
           QMouseEvent.modifiers(): Qt::KeyboardModifiers
           QFlags<KeyboardModifier>
             Qt::NoModifier          empty string
             Qt::ShiftModifier       Shft
             Qt::ControlModifier     Ctrl
             Qt::AltModifier         Alt
             Qt::MetaModifier        Meta
             Qt::KeypadModifier      Keypad: unused
             Qt::GroupSwitchModifier unused
        - Mouse events
             MouseButtonPress        Press Left/Right/Middle
             MouseMove               Move
             MouseButtonRelease      Release
             MouseButtonDblClick     DblClick
             Wheel                   Wheel
               event is separated on Qt but processes here
        - Position
             Inside one of the available positions
    """
    def __init__(self, widget: T):
        self._widget : T = widget
        # Instance of motion action
        self._current_motion    : Optional[MouseMotionActions] = None

        # Set key events callbacks
        # Each event will be associate with a unique string
        self._mouse_callback : Dict[str, Callable             ] = {}
        self._motion_classes : Dict[str, Type[MouseMotionActions[T]]] = {}

        # Regions to process events differently
        # A function which returns a bool is given to check if the 
        # current mouse position is inside the region
        self._regions : Dict[str, Callable[[QtCore.QPoint], bool]] = {}

    def _get_markdown_help(self) -> str:
        res = ''
        res += '|Mouse    |Action  |  \n'
        res += '|:--------|:------:|  \n'
        # TODO create html table
        for k,v in self._mouse_callback.items():
            res += f'|{k}|{v.__doc__}|  \n'
        for k,v in self._motion_classes.items():
            res += f'|{k}|{v.__doc__}|  \n'
        res += '  \n'
        return res

    def markdown_help(self) -> str:
        """ return events documentation in markdown format """
        return self._get_markdown_help()

    def modifiers2str(self, modifiers: QtCore.Qt.KeyboardModifiers ) -> str:
        """ Converts the modifiers to a string representation """
        res = ''
        qt_mod = QtCore.Qt.KeyboardModifiers
        mod_str = {
            qt_mod.ShiftModifier   : 'Shft',
            qt_mod.ControlModifier : 'Ctrl',
            qt_mod.AltModifier     : 'Alt',
            qt_mod.MetaModifier    : 'Meta',
        }
        for m,m_str in mod_str.items():
            if modifiers & m:
                res = add2repr(res, m_str)
        return res

    def buttons2str(self, buttons: QtGui.Qt.MouseButtons ) -> str:
        """ Converts the mouse buttons to a string representation """
        res = ''
        qt_but = QtGui.Qt.MouseButton
        button_str = {
            qt_but.LeftButton   : 'Left',
            qt_but.RightButton  : 'Right',
            qt_but.MiddleButton : 'Middle',
        }
        for b,b_str in button_str.items():
            if buttons & b:
                res = add2repr(res, b_str)
        return res

    def process_event(self, event_repr: str, event : QtCore.QEvent) -> bool:
        """ Process the event base on its representation as string and on the dict of callbacks """
        if event_repr in self._mouse_callback:
            return self._mouse_callback[event_repr](event)
        return False

    def start_motion(self, event_repr: str, event : QtGui.QMouseEvent) -> bool:
        """ Instantiate motion actions """
        if event_repr in self._motion_classes and self._current_motion is None:
            self._current_motion = self._motion_classes[event_repr](self._widget)
            self._current_motion.press(event)
            return True
        return False

    def mouse_press_event(self, event : QtGui.QMouseEvent):
        """ Build str that represents the event and call process_event() """
        event_repr : str = ''
        # 1. Get Modifiers
        event_repr = add2repr(event_repr, self.modifiers2str(event.modifiers()))
        # 2. Get Buttons
        event_repr = add2repr(event_repr, self.buttons2str(event.buttons()))
        motion_event = event_repr + ' Motion'
        press_event  = event_repr + ' Pressed'
        processed = self.process_event(press_event,  event)
        self.start_motion (motion_event, event)
        # Propagate event to parent anyway?
        event.setAccepted(processed)

    def mouse_release_event(self, event: QtGui.QMouseEvent) -> None:
        """ Build str that represents the event and call process_event() """
        if self._current_motion:
            self._current_motion.release(event)
            self._current_motion = None
            event.accept()
        event.ignore()

    def mouse_move_event(self, event: QtGui.QMoveEvent):
        """ Build str that represents the event and call process_event() """
        # self._viewer.mouse_pos = event.pos()
        # We save the event type in a member variable to be able to process the release event
        if self._current_motion:
            self._current_motion.move(event)
        else:
            event_repr : str = 'Move'
            event_repr += self.check_regions(event.pos())
            processed = self.process_event(event_repr,  event)
            if processed:
                event.setAccepted(True)
            else:
                self.mouse_move_unpressed(event)

    @abstractmethod
    def mouse_move_unpressed(self, event: QtGui.QMoveEvent)->None:
        """ Actions while moving the mouse without pressing any button """

    def check_regions(self, pos: QtCore.QPoint )->str:
        """ Return region information for event representation string """
        res = ''
        for name,call in self._regions.items():
            if call(pos):
                res += f' on {name}'
        return res

    def mouse_double_click_event(self, event: QtGui.QMouseEvent):
        """ Deal with double-click event """
        event_repr : str = ''
        # 1. Get Modifiers
        event_repr = add2repr(event_repr, self.modifiers2str(event.modifiers()))
        # 2. Get Buttons
        event_repr = add2repr(event_repr, self.buttons2str(event.buttons()))
        event_repr = add2repr(event_repr, 'DblClick')
        # Check if double click is on any defined region
        event_repr += self.check_regions(event.pos())
        print(f"{event_repr}")
        processed = self.process_event(event_repr,  event)
        event.setAccepted(processed)

    def mouse_wheel_event(self,event: QtGui.QWheelEvent) -> None:
        """ Build str that represents the wheek event and call process_event() """
        # Can add button states
        # 1. Get Modifiers
        event_repr = add2repr('', self.modifiers2str(event.modifiers()))
        event_repr = add2repr(event_repr, 'Wheel')
        # print(f"{event_repr}")
        processed = self.process_event(event_repr,  event)
        event.setAccepted(processed)




from PySide6 import QtGui, QtWidgets, QtCore, QtMultimedia, QtMultimediaWidgets
from PySide6.QtOpenGLWidgets import QOpenGLWidget as QOpenGLWidget
from PySide6.QtCore import Signal, Slot

QLabel = QtWidgets.QLabel
QApplication = QtWidgets.QApplication
QTimer = QtCore.QTimer




from _md5 import md5
import numpy as np
import os
from typing import List
from ..utils.qt_imports import QtGui, QtCore

# Adapt to different versions
qevent_types = QtCore.QEvent.Type

MOUSE_EVENTS : List[qevent_types] = [
            # qevent_typese.KeyPress,
            qevent_types.MouseButtonPress,
            qevent_types.MouseButtonRelease,
            qevent_types.MouseMove,
            qevent_types.MouseButtonDblClick,
        ]

RESIZE_EVENTS : List[qevent_types] = [ qevent_types.Resize ]

WHEEL_EVENT : List[qevent_types] = [ qevent_types.Wheel]

class QtDump:

    # ----- Qt enum value -----
    @staticmethod
    def enum_value(v):
        return v.value

    # ----- Convert QMouseEvent -----
    @staticmethod
    def mouseevent2dict(evt):
        evt_type = evt.type()
        button = evt.button()
        buttons = evt.buttons()
        pos = evt.localPos()
        modifiers = evt.modifiers()
        # Transform each information
        info = {
            'type':     QtDump.enum_value(evt_type),
            'button':   QtDump.enum_value(button),
            'buttons':  QtDump.enum_value(buttons),
            'pos': [pos.x(), pos.y()],
            'modifiers': QtDump.enum_value(modifiers)
        }
        return info

    @staticmethod
    def dict2mouseevent(info):
        pos = QtCore.QPointF(info['pos'][0], info['pos'][1])
        mouse_event = QtGui.QMouseEvent(
            QtCore.QEvent.Type(info['type']),
            pos,
            QtCore.Qt.MouseButton(info['button']),
            QtCore.Qt.MouseButtons(info['buttons']),
            QtCore.Qt.KeyboardModifiers(info['modifiers']))
        return mouse_event

    # ----- Convert QSize -----
    @staticmethod
    def size2dict(size):
        width = size.width()
        height = size.height()
        # Transform each information
        info = {
            'width': width,
            'height': height,
        }
        return info

    @staticmethod
    def dict2size(info):
        return QtCore.QSize(info['width'], info['height'])

    # ----- Convert QPointF -----
    @staticmethod
    def qpointf2dict(pt):
        x = pt.x()
        y = pt.y()
        # Transform each information
        info = {
            'x': x,
            'y': y,
        }
        return info

    @staticmethod
    def dict2qpointf(info):
        return QtCore.QPointF(info['x'], info['y'])

    # ----- Convert QPoint -----
    @staticmethod
    def qpoint2dict(pt):
        x = pt.x()
        y = pt.y()
        # Transform each information
        info = {
            'x': x,
            'y': y,
        }
        return info

    @staticmethod
    def dict2qpoint(info):
        return QtCore.QPoint(info['x'], info['y'])

    # ----- Convert QWheelEvent -----
    @staticmethod
    def wheelevent2dict(evt):
        evt_type = evt.type()
        # Transform each information
        info = {
            'type':       evt_type.__int__(),
            'pos':        QtDump.qpointf2dict(evt.position()),
            'globalPos':  QtDump.qpointf2dict(evt.globalPosition()),
            'pixelDelta': QtDump.qpoint2dict (evt.pixelDelta()),
            'angleDelta': QtDump.qpoint2dict (evt.angleDelta()),
            'buttons':    QtDump.enum_value(evt.buttons()),
            'modifiers':  QtDump.enum_value(evt.modifiers()),
            'phase':      QtDump.enum_value(evt.phase()),
            'inverted':   evt.inverted(),
        }
        return info

    @staticmethod
    def dict2wheelevent(info):
        pos        = QtDump.dict2qpointf(         info['pos'])
        globalPos  = QtDump.dict2qpointf(         info['globalPos'])
        pixelDelta = QtDump.dict2qpoint(          info['pixelDelta'])
        angleDelta = QtDump.dict2qpoint(          info['angleDelta'])
        buttons    = QtCore.Qt.MouseButtons(      info['buttons'])
        modifiers  = QtCore.Qt.KeyboardModifiers( info['modifiers'])
        phase      = QtCore.Qt.ScrollPhase(       info['phase'])
        inverted   = info['inverted']
        event = QtGui.QWheelEvent(pos, globalPos, pixelDelta, angleDelta, buttons, modifiers, phase, inverted)
        return event

    # ----- Convert QResizeEvent -----
    @staticmethod
    def resizeevent2dict(evt):
        evt_type = evt.type()
        # Transform each information
        info = {
            'type': evt_type.__int__(),
            'size':    QtDump.size2dict(evt.size()),
            'oldSize': QtDump.size2dict(evt.oldSize()),
        }
        return info

    @staticmethod
    def dict2resizeevent(info):
        size = QtDump.dict2size(info['size'])
        oldSize = QtDump.dict2size(info['oldSize'])
        resize_event = QtGui.QResizeEvent(size, oldSize)
        return resize_event

    # ----- Convert dict to event -----
    @staticmethod
    def dict2event(info):
        if info['type'] in MOUSE_EVENTS:
            # print(f"mouse_event {info['type']}")
            return QtDump.dict2mouseevent(info)
        else:
            if info['type'] in RESIZE_EVENTS:
                return QtDump.dict2resizeevent(info)
            else:
                if info['type'] in WHEEL_EVENT:
                    return QtDump.dict2wheelevent(info)
                else:
                    print("Error: event not available")

    # ----- widget snapshot hash -----
    @staticmethod
    def get_screen_hash(widget):
        image = QtGui.QImage(widget.size(), QtGui.QImage.Format.Format_RGB32)
        widget.render(image)
        i = 0
        while os.path.isfile(f"_display_dump_{i}.png"):
            i += 1
        image.save(f"_display_dump_{i}.png")
        # get hash from image data pixmap.toImage().bits()
        image_bits = image.constBits()
        arr = np.array(image_bits).reshape(image.height(), image.width(), 4)
        image_hash = md5(arr).hexdigest()
        return image_hash



"""
    MultiView class: display and compare multiple images
"""
import math
import types
from enum                         import Enum, auto
from typing                       import List, Optional, NewType
from qimview.utils.qt_imports     import QtGui, QtWidgets, QtCore
from qimview.utils.utils          import get_time
from qimview.utils.viewer_image   import ImageFormat
from qimview.utils.menu_selection import MenuSelection
from qimview.utils.mvlabel        import MVLabel
from qimview.cache                import ImageCache
from qimview.image_viewers        import (QTImageViewer, GLImageViewer, GLImageViewerShaders,
                                          ImageFilterParameters, ImageFilterParametersGui)
from qimview.image_viewers.image_viewer import ImageViewer
from .fullscreen_helper                 import FullScreenHelper
from .multi_view_key_events             import MultiViewKeyEvents
from .multi_view_mouse_events           import MultiViewMouseEvents

# Class that derives from ImageViewer
ImageViewerClass = NewType('ImageViewerClass', ImageViewer)

class ViewerType(Enum):
    """ Different flavors of ImageViewer """
    QT_VIEWER             = auto()
    OPENGL_VIEWER         = auto()
    OPENGL_SHADERS_VIEWER = auto()


class MultiView(QtWidgets.QWidget):
    """ Multiple image viewer: hold a grid layout of several image viewers and a set of input images.
        Each viewer has the possibility to display any of the input images.
        Features:
          - geometric and intensity comparison by synchronizing the different viewers.
          - synchronized image filters
          - fast switch between images
    """

    def __init__(self, parent=None, viewer_mode: ViewerType =ViewerType.QT_VIEWER, nb_viewers: int =1) -> None:
        """
        :param parent:
        :param viewer_mode:
        :param nb_viewers_used:
        """
        QtWidgets.QWidget.__init__(self, parent)

        # --- Protected members
        # Active viewer index, -1 if none viewer is active
        self._active_viewer_index  : int                        = -1
        self._active_viewer        : Optional[ImageViewerClass] = None
        # Show only active window
        self._show_active_only     : bool                       = False

        # FullScreen helper features
        self._fullscreen           : FullScreenHelper           = FullScreenHelper()

        # Key event class
        self._key_events   : MultiViewKeyEvents   = MultiViewKeyEvents(self)
        # Mouse event class
        self._mouse_events : MultiViewMouseEvents = MultiViewMouseEvents(self)

        # Clipboard
        self._save_image_clipboard : bool                       = False
        self._clipboard            : Optional[QtGui.QClipboard] = None

        # --- Public members
        self.use_opengl = viewer_mode in [ViewerType.OPENGL_SHADERS_VIEWER, ViewerType.OPENGL_VIEWER]

        self.nb_viewers_used : int = nb_viewers
        self.allocated_image_viewers = []  # keep allocated image viewers here
        self.image_viewer_classes = {
            ViewerType.QT_VIEWER:             QTImageViewer,
            ViewerType.OPENGL_VIEWER:         GLImageViewer,
            ViewerType.OPENGL_SHADERS_VIEWER: GLImageViewerShaders
        }
        self.image_viewer_class = self.image_viewer_classes[viewer_mode]

        self.image_viewers : List[ImageViewerClass] = []
        # Create viewer instances
        for n in range(self.nb_viewers_used):
            viewer = self.create_viewer()
            self.allocated_image_viewers.append(viewer)
            self.image_viewers.append(viewer)

        self.viewer_mode = viewer_mode
        self.bold_font = QtGui.QFont()

        self.verbosity_LIGHT = 1
        self.verbosity_TIMING = 1 << 2
        self.verbosity_TIMING_DETAILED = 1 << 3
        self.verbosity_TRACE = 1 << 4
        self.verbosity_DEBUG = 1 << 5
        self.verbosity = 0

        # self.set_verbosity(self.verbosity_LIGHT)
        # self.set_verbosity(self.verbosity_TIMING_DETAILED)
        # self.set_verbosity(self.verbosity_TRACE)

        self.current_image_filename = None

        self.filter_params = ImageFilterParameters()
        self.filter_params_gui = ImageFilterParametersGui(self.filter_params)

        self.raw_bayer = {
            'Read': None, 
            'Bayer0': ImageFormat.CH_GBRG, 
            'Bayer1': ImageFormat.CH_BGGR, 
            'Bayer2': ImageFormat.CH_RGGB, 
            'Bayer3': ImageFormat.CH_GRBG
        }
        self.default_raw_bayer = 'Read'
        self.current_raw_bayer = self.default_raw_bayer

        # Number of viewers currently displayed
        self.nb_viewers_used : int = 0

        # save images of last visited row
        self.cache = ImageCache()
        self.image_dict = { }
        self.read_size = 'full'
        self.image1 = dict()
        self.image2 = dict()
        self.button_layout = None
        self.message_cb = None
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

        self.key_up_callback = None
        self.key_down_callback = None
        self.output_image_label = dict()

        self.output_label_current_image   : str = ''
        self.output_label_reference_image : str = ''
        self.add_context_menu()
        
        # Parameter to set the number of columns in the viewer grid layout
        # if 0: computed automatically
        self.max_columns       : int = 0 

    def create_viewer(self) -> ImageViewerClass:
        viewer = self.image_viewer_class(self)
        viewer.set_activation_callback(self.on_active)
        viewer.set_synchronization_callback(self.on_synchronize)
        viewer.add_help_tab('MultiView keys',  self._key_events  .markdown_help())
        viewer.add_help_tab('MultiView Mouse', self._mouse_events.markdown_help())
        viewer.add_help_links(self._key_events.help_links())
        return viewer

    def set_key_up_callback(self, c):
        self.key_up_callback = c

    def set_key_down_callback(self, c):
        self.key_down_callback = c

    def add_context_menu(self):
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.show_context_menu)
        self._context_menu = QtWidgets.QMenu()
        self.viewer_modes = { v.name:v for v in ViewerType}
        self._default_viewer_mode = ViewerType.QT_VIEWER.name
        self.viewer_mode_selection = MenuSelection("Viewer mode", 
            self._context_menu, self.viewer_modes, self._default_viewer_mode, self.update_viewer_mode)
        self._context_menu.addSeparator()
        action = self._context_menu.addAction("Reset viewers")
        action.triggered.connect(self.reset_viewers)

    def reset_viewers(self):
        for v in self.image_viewers:
            v.widget.hide()
            self.viewer_grid_layout.removeWidget(v.widget)
        self.allocated_image_viewers.clear()
        self.image_viewers.clear()
        # Create viewer instances
        for n in range(self.nb_viewers_used):
            viewer = self.create_viewer()
            # viewer.setContextMenuPolicy(QtCore.Qt.PreventContextMenu)
            self.allocated_image_viewers.append(viewer)
            self.image_viewers.append(viewer)
        self.set_number_of_viewers(self.nb_viewers_used)
        self.viewer_grid_layout.update()
        self.update_image()

    def update_viewer_mode(self):
        viewer_mode = self.viewer_mode_selection.get_selection_value()
        self.image_viewer_class = self.image_viewer_classes[viewer_mode]

    def show_context_menu(self, pos):
        self._context_menu.show()
        self._context_menu.popup( self.mapToGlobal(pos) )

    def set_cache_memory_bar(self, progress_bar):
        self.cache.set_memory_bar(progress_bar)

    def set_verbosity(self, flag, enable=True):
        """
        :param v: verbosity flags
        :param b: boolean to enable or disable flag
        :return:
        """
        if enable:
            self.verbosity = self.verbosity | flag
        else:
            self.verbosity = self.verbosity & ~flag

    def check_verbosity(self, flag) -> int:
        return self.verbosity & flag

    def print_log(self, mess):
        if self.verbosity & self.verbosity_LIGHT:
            print(mess)

    def show_timing(self) -> int:
        return self.check_verbosity(self.verbosity_TIMING) or self.check_verbosity(self.verbosity_TIMING_DETAILED)

    def show_timing_detailed(self) -> bool:
        return (self.check_verbosity(self.verbosity_TIMING_DETAILED)>0)

    def show_trace(self):
        return self.check_verbosity(self.verbosity_TRACE)

    def make_mouse_press(self, image_name):
        def mouse_press(obj, event):
            print('mouse_press')
            obj.update_image(image_name)

        return types.MethodType(mouse_press, self)

    def mouse_release(self, event):
        self.update_image(self.output_label_reference_image)

    def make_mouse_double_click(self, image_name):
        def mouse_double_click(obj, event):
            '''
            Sets the double clicked label as the reference image
            :param obj:
            :param event:
            '''
            print('mouse_double_click {}'.format(image_name))
            obj.output_label_reference_image = image_name
            obj.output_label_current_image = obj.output_label_reference_image
            obj.update_image()

        return types.MethodType(mouse_double_click, self)

    def set_read_size(self, read_size):
        self.read_size = read_size
        # reset cache
        self.cache.reset()

    def update_image_intensity_event(self):
        self.update_image_parameters()

    def reset_intensities(self):
        self.filter_params_gui.reset_all()

    def update_image_parameters(self):
        '''
        Uses the variable self.output_label_current_image
        :return:
        '''
        self.print_log('update_image_parameters')
        update_start = get_time()

        for n in range(self.nb_viewers_used):
            self.image_viewers[n].filter_params.copy_from(self.filter_params)
            self.image_viewers[n].widget.update()

        if self.show_timing():
            time_spent = get_time() - update_start
            self.print_log(" Update image took {0:0.3f} sec.".format(time_spent))

    def set_images(self, images, set_viewers=False):
        self.print_log(f"MultiView.set_images() {images}")
        if images.keys() == self.image_dict.keys():
            self.image_dict = images
            self.update_reference()
        else:
            self.image_dict = images
            self.update_image_buttons()

    def set_viewer_images(self):
        """
        Set viewer images based on self.image_dict.keys()
        :return:
        """
        # if set_viewers, we force the viewer layout and images based on the list
        # be sure to have enough image viewers allocated
        while self.nb_viewers_used > len(self.allocated_image_viewers):
            viewer = self.create_viewer()
            self.allocated_image_viewers.append(viewer)
        self.image_viewers = self.allocated_image_viewers[:self.nb_viewers_used]
        image_names = list(self.image_dict.keys())
        for n in range(self.nb_viewers_used):
            self.image_viewers[n].image_name = image_names[min(n,len(image_names)-1)]

    def update_reference(self) -> None:
        reference_image = self.get_output_image(self.output_label_reference_image)
        for n in range(self.nb_viewers_used):
            viewer = self.image_viewers[n]
            # set reference image
            viewer.set_image_ref(reference_image)
            viewer.image_ref_name = self.output_label_reference_image

    def set_reference_label(self, ref: str, update_viewers=False) -> None:
        try:
            if ref is not None:
                if ref!=self.output_label_reference_image:
                    self.output_label_reference_image = ref
                    if update_viewers:
                        self.update_reference()
        except Exception as e:
            print(f' Failed to set reference label {e}')

    def update_image_buttons(self):
        # choose image to display
        self.clear_buttons()
        self.image_list = list(self.image_dict.keys())
        self.print_log("MultiView.update_image_buttons() {}".format(self.image_list))
        self.label = dict()
        for image_name in self.image_list:
            # possibility to disable an image using the string 'none', especially useful for input image
            if image_name != 'none':
                self.label[image_name] = MVLabel(image_name, self)
                self.label[image_name].setFrameShape(QtWidgets.QFrame.Shape.Panel)
                self.label[image_name].setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
                # self.label[image_name].setLineWidth(3)
                self.label[image_name].setSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, 
                                                     QtWidgets.QSizePolicy.Policy.Minimum)
                # self.label[image_name].setFixedHeight(40)
                self.label[image_name].mousePressEvent = self.make_mouse_press(image_name)
                self.label[image_name].mouseReleaseEvent = self.mouse_release
                self.label[image_name].mouseDoubleClickEvent = self.make_mouse_double_click(image_name)
        self.create_buttons()

        # the crop area can be changed using the mouse wheel
        self.output_label_crop = (0., 0., 1., 1.)

        if len(self.image_list)>0:
            self.output_label_current_image = self.image_list[0]
            self.set_reference_label(self.image_list[0], update_viewers=True)
        else:
            self.output_label_current_image = ''
            self.output_label_reference_image = ''

    def clear_buttons(self):
        if self.button_layout is not None:
            # start clearing the layout
            # for i in range(self.button_layout.count()): self.button_layout.itemAt(i).widget().close()
            self.print_log(f"MultiView.clear_buttons() {self.image_list}")
            for image_name in reversed(self.image_list):
                if image_name in self.label:
                    self.button_layout.removeWidget(self.label[image_name])
                    self.label[image_name].close()

    def create_buttons(self):
        if self.button_layout is not None:
            max_grid_columns = 10
            idx = 0
            for image_name in self.image_list:
                # possibility to disable an image using the string 'none', especially useful for input image
                if image_name != 'none':
                    self.button_layout.addWidget(self.label[image_name], idx // max_grid_columns, idx % max_grid_columns)
                    idx += 1

    def layout_buttons(self, vertical_layout):
        self.button_widget = QtWidgets.QWidget(self)
        self.button_layout = QtWidgets.QGridLayout()
        self.button_layout.setHorizontalSpacing(0)
        self.button_layout.setVerticalSpacing(0)
        # button_layout.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.create_buttons()
        vertical_layout.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinimumSize)
        # vertical_layout.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.button_widget.setSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
        self.button_widget.setLayout(self.button_layout)
        vertical_layout.addWidget(self.button_widget, 0, QtCore.Qt.AlignmentFlag.AlignTop)

    def layout_parameters(self, parameters_layout):
        # Add Profiles and keep zoom options
        self.display_profiles = QtWidgets.QCheckBox("Profiles")
        self.display_profiles.stateChanged.connect(self.toggle_display_profiles)
        self.display_profiles.setChecked(False)
        parameters_layout.addWidget(self.display_profiles)
        self.keep_zoom = QtWidgets.QCheckBox("Keep zoom")
        self.keep_zoom.setChecked(False)
        parameters_layout.addWidget(self.keep_zoom)

        # Reset button
        self.reset_button = QtWidgets.QPushButton("reset")
        parameters_layout.addWidget(self.reset_button)
        self.reset_button.clicked.connect(self.reset_intensities)

        # Add color difference slider
        self.filter_params_gui.add_imdiff_factor(parameters_layout, self.update_image_intensity_event)

        # --- Saturation adjustment
        self.filter_params_gui.add_saturation(parameters_layout, self.update_image_intensity_event)
        # --- Black point adjustment
        self.filter_params_gui.add_blackpoint(parameters_layout, self.update_image_intensity_event)
        # --- white point adjustment
        self.filter_params_gui.add_whitepoint(parameters_layout, self.update_image_intensity_event)
        # --- Gamma adjustment
        self.filter_params_gui.add_gamma(parameters_layout, self.update_image_intensity_event)

    def layout_parameters_2(self, parameters2_layout):
        # --- G_R adjustment
        self.filter_params_gui.add_g_r(parameters2_layout, self.update_image_intensity_event)
        # --- G_B adjustment
        self.filter_params_gui.add_g_b(parameters2_layout, self.update_image_intensity_event)

    def update_layout(self):
        self.print_log("update_layout")
        vertical_layout = QtWidgets.QVBoxLayout()
        self.layout_buttons(vertical_layout)

        # First line of parameter control
        parameters_layout = QtWidgets.QHBoxLayout()
        self.layout_parameters(parameters_layout)
        vertical_layout.addLayout(parameters_layout, 1)

        # Second line of parameter control
        parameters2_layout = QtWidgets.QHBoxLayout()
        self.layout_parameters_2(parameters2_layout)
        vertical_layout.addLayout(parameters2_layout, 1)

        self.viewer_grid_layout = QtWidgets.QGridLayout()
        self.viewer_grid_layout.setHorizontalSpacing(1)
        self.viewer_grid_layout.setVerticalSpacing(1)
        self.set_number_of_viewers(1)
        vertical_layout.addLayout(self.viewer_grid_layout, 1)

        self.figures_widget = QtWidgets.QWidget()
        self.figures_layout = QtWidgets.QHBoxLayout()
        self.figures_layout.setSizeConstraint(QtWidgets.QLayout.SizeConstraint.SetMinimumSize)
        # for the moment ignore this
        # self.figures_layout.addWidget(self.value_in_range_canvas)
        # self.figures_widget.setLayout(self.figures_layout)

        vertical_layout.addWidget(self.figures_widget)
        self.toggle_display_profiles()
        self.setLayout(vertical_layout)
        print("update_layout done")

    def toggle_display_profiles(self):
        self.figures_widget.setVisible(self.display_profiles.isChecked())
        self.update_image()

    def get_output_image(self, im_string_id : str):
        """
        Search for the image with given label in the current row
        if not in cache reads it and add it to the cache
        :param im_string_id: string that identifies the image to display
        :return:
        """
        # print(f"get_output_image({im_string_id}) ")
        start = get_time()

        image_filename = self.image_dict[im_string_id]
        image_transform = None
        self.print_log(f"MultiView.get_output_image() image_filename:{image_filename}")

        image_data, _ = self.cache.get_image(image_filename, self.read_size, verbose=self.show_timing_detailed(),
                                             use_RGB=not self.use_opengl, image_transform=image_transform)

        if image_data is not None:
            self.output_image_label[im_string_id] = image_filename
            output_image = image_data
        else:
            print(f"failed to get image {im_string_id}: {image_filename}")
            return None

        if self.show_timing_detailed():
            print(f" get_output_image took {int((get_time() - start)*1000+0.5)} ms".format)

        # force image bayer information if selected from menu
        res = output_image
        set_bayer = self.raw_bayer[self.current_raw_bayer]
        if res.channels in [ImageFormat.CH_BGGR, ImageFormat.CH_GBRG, ImageFormat.CH_GRBG, ImageFormat.CH_RGGB] and set_bayer is not None:
            print(f"Setting bayer {set_bayer}")
            res.channels = set_bayer

        return res

    def set_message_callback(self, message_cb):
        self.message_cb = message_cb

    def setMessage(self, mess):
        if self.message_cb is not None:
            self.message_cb(mess)

    def cache_read_images(self, image_filenames: List[str], reload: bool =False) -> None:
        """ Read the list of images into the cache, with option to reload them from disk

        Args:
            image_filenames (List[str]): list of image filenames
            reload (bool, optional): reload removes first the images from the ImageCache 
                before adding them. Defaults to False.
        """        
        # print(f"cache_read_images({image_filenames}) ")
        image_transform = None
        if reload:
            for f in image_filenames:
                self.cache.remove(f)
        self.cache.add_images(image_filenames, self.read_size, verbose=False, use_RGB=not self.use_opengl,
                             image_transform=image_transform)

    def update_label_fonts(self):
        # Update selected image label, we could do it later too
        for im_name in self.image_list:
            # possibility to disable an image using the string 'none', especially useful for input image
            if im_name != 'none':
                is_bold      = im_name == self.output_label_current_image
                is_underline = im_name == self.output_label_reference_image
                is_bold |= is_underline
                self.bold_font.setBold(is_bold)
                self.bold_font.setUnderline(is_underline)
                self.bold_font.setPointSize(8)
                self.label[im_name].setFont(self.bold_font)
                self.label[im_name].setWordWrap(True)
            # self.label[im_name].setMaximumWidth(160)
    
    def on_active(self, viewer : ImageViewerClass) -> None:
        # Activation requested for a given viewer
        self._active_viewer = viewer
        self.update_image(viewer.image_name)
 
    def on_synchronize(self, viewer : ImageViewerClass) -> None:
        # Synchronize other viewer to calling viewer
        for v in self.image_viewers:
            if v != viewer:
                viewer.synchronize_data(v)
                v.viewer_update()
                # Force immediate paint
                # viewer.repaint()
    
    def set_clipboard(self, clipboard : Optional[QtGui.QClipboard], save_image: bool):
        self._clipboard            = clipboard
        self._save_image_clipboard = save_image

    def update_image(self, image_name=None, reload=False):
        """
        Uses the variable self.output_label_current_image
        :return:
        """
        self.print_log('update_image {} current: {}'.format(image_name, self.output_label_current_image))
        update_image_start = get_time()

        # Checks
        # Define the current selected image
        if image_name is not None: self.output_label_current_image = image_name
        if self.output_label_current_image == "" or len(self.image_viewers) == 0: return

        if self.image_dict[self.output_label_current_image] is None:
            print(" No image filename for current image")
            return

        self.update_label_fonts()

        # If not found, set to 0
        if self._active_viewer not in self.image_viewers: 
            self._active_viewer = self.image_viewers[0]
        self._active_viewer.display_timing = self.show_timing()>0

        # Read images in parallel to improve preformances
        # list all required image filenames
        # set all viewers image names (labels)
        image_filenames = [self.image_dict[self.output_label_current_image]]
        # define image associated to each used viewer and add it to the list of images to get
        for n in range(self.nb_viewers_used):
            viewer : ImageViewer = self.image_viewers[n]
            # Set active only the first active window
            viewer.is_active = (viewer == self._active_viewer)
            if viewer.get_image() is None:
                if n < len(self.image_list):
                    viewer.image_name = self.image_list[n]
                    image_filenames.append(self.image_dict[self.image_list[n]])
                else:
                    viewer.image_name = self.output_label_current_image
            else:
                # image_name should belong to image_dict
                if viewer.image_name in self.image_dict:
                    image_filenames.append(self.image_dict[viewer.image_name])
                else:
                    viewer.image_name = self.output_label_current_image

        # remove duplicates
        image_filenames = list(set(image_filenames))
        # print(f"image filenames {image_filenames}")
        self.cache_read_images(image_filenames, reload=reload)

        try:
            current_image = self.get_output_image(self.output_label_current_image)
            if current_image is None: return
        except Exception as e:
            print("Error: failed to get image {}: {}".format(self.output_label_current_image, e))
            return

        # print(f"cur {self.output_label_current_image}")
        current_filename = self.output_image_label[self.output_label_current_image]

        # if self.show_timing_detailed(): time_spent = get_time() - update_image_start

        self.setMessage("Image: {0}".format(current_filename))

        current_viewer = self._active_viewer
        current_viewer.is_active = True
        current_viewer.image_name = self.output_label_current_image
        current_viewer.set_image(current_image)

        # print(f"ref {self.output_label_reference_image}")
        if self.output_label_reference_image==self.output_label_current_image:
            reference_image = current_image
        else:
            reference_image = self.get_output_image(self.output_label_reference_image)

        if self.nb_viewers_used >= 2:
            for n in range(self.nb_viewers_used):
                viewer : ImageViewer = self.image_viewers[n]
                if viewer != self._active_viewer:
                    # Update viewer images
                    try:
                        viewer_image = self.get_output_image(viewer.image_name)
                    except Exception as e:
                        print("Error: failed to get image {}: {}".format(viewer.image_name, e))
                        viewer_image = current_image
                    # Be sure to update image data
                    viewer.set_image(viewer_image)
                    # set reference image
                    viewer.set_image_ref(reference_image)
                    viewer.image_ref_name = self.output_label_reference_image

        # if self._save_image_clipboard and self._clipboard:
        #     print("set save image to clipboard")
        #     self._active_viewer.set_clipboard(self._clipboard, True)

        # Be sure to show the required viewers
        if self._show_active_only:
            for n in range(self.nb_viewers_used):
                viewer = self.image_viewers[n]
                viewer.widget.setVisible(viewer.is_active)
            self._active_viewer.widget.update()
        else:
            for n in range(self.nb_viewers_used):
                viewer = self.image_viewers[n]
                # Note: calling show in any case seems to avoid double calls to paint event that update() triggers
                # viewer.show()
                if viewer.widget.isHidden(): viewer.widget.show()
                else:                        viewer.widget.update()

        if self._save_image_clipboard:
            print("set save image to clipboard")
            self._active_viewer.set_clipboard(self._clipboard, True)
            self._active_viewer.widget.repaint()
            print("end save image to clipboard")
            self._active_viewer.set_clipboard(None, False)

        # self.image_scroll_area.adjustSize()
        # if self.show_timing():
        print(f" Update image took {(get_time() - update_image_start)*1000:0.0f} ms")

    def set_number_of_viewers(self, nb_viewers: int = 1, max_columns : int = 0) -> None:
        self.print_log("*** set_number_of_viewers()")
        if nb_viewers<1: return

        # 1. remove current viewers from grid layout
        # self.viewer_grid_layout.hide()
        for v in self.image_viewers:
            v.widget.hide()
            self.viewer_grid_layout.removeWidget(v)

        self.nb_viewers_used : int = nb_viewers
        print(f"max_columns = {max_columns}")
        if max_columns>0:
            row_length = min(self.nb_viewers_used, max_columns)
            col_length = int(math.ceil(self.nb_viewers_used / row_length))
        else:
            # Find best configuration to fill the space based on image size and widget size?
            col_length = int(math.sqrt(self.nb_viewers_used)+0.5)
            row_length = int(math.ceil(self.nb_viewers_used / col_length))
        self.print_log('col_length = {} row_length = {}'.format(col_length, row_length))
        # be sure to have enough image viewers allocated
        while self.nb_viewers_used > len(self.allocated_image_viewers):
            viewer = self.create_viewer()
            self.allocated_image_viewers.append(viewer)

        self.image_viewers = self.allocated_image_viewers[:self.nb_viewers_used]

        for n in range(self.nb_viewers_used):
            self.viewer_grid_layout.addWidget(self.image_viewers[n], int(n / float(row_length)), n % row_length)
            self.image_viewers[n].hide()
        
        # Update image parameters for all viewers
        self.update_image_parameters()
        # for n in range(self.nb_viewers_used):
        #     print("Viewer {} size {}".format(n, (self.image_viewers[n].width(), self.image_viewers[n].height())))

    def mouseDoubleClickEvent(self, event):
        self._mouse_events.mouse_double_click_event(event)

    def mousePressEvent(self, event):
        self._mouse_events.mouse_press_event(event)

    def keyPressEvent(self, event):
        self._key_events.key_press_event(event)



#
#
# started from https://cyrille.rossant.net/2d-graphics-rendering-tutorial-with-pyopengl/
#
#

from qimview.utils.qt_imports import QtWidgets
from .image_viewer            import trace_method, get_time
from .gl_image_viewer_base    import GLImageViewerBase

from PySide6.QtOpenGL import (QOpenGLBuffer, QOpenGLShader,
                              QOpenGLShaderProgram, QOpenGLTexture)

import OpenGL.GL as gl
from OpenGL.GL import shaders
import argparse
import sys
import numpy as np

class GLImageViewerShaders(GLImageViewerBase):
    # vertex shader program
    vertexShader = """
        #version 330 core

        attribute vec3 vert;
        attribute vec2 uV;
        uniform mat4 mvMatrix;
        uniform mat4 pMatrix;
        out vec2 UV;

        void main() {
          gl_Position = pMatrix * mvMatrix * vec4(vert, 1.0);
          UV = uV;
        }
        """
    # fragment shader program
    fragmentShader_RGB = """
        #version 330 core
    
        in vec2 UV;
        uniform sampler2D backgroundTexture;
        uniform int channels; // channel representation
        uniform float white_level;
        uniform float black_level;
        uniform float g_r_coeff;
        uniform float g_b_coeff;
        uniform float max_value; // maximal value based on image precision
        uniform float max_type;  // maximal value based on image type (uint8, etc...)
        uniform float gamma;
        out vec3 colour;
    
        void main() {
          colour = texture(backgroundTexture, UV).rgb;

          // black level
          colour.rgb = colour.rgb/max_value*max_type;
          colour.rgb = max((colour.rgb-vec3(black_level).rgb),0);

          // white balance
          colour.r = colour.r*g_r_coeff;
          colour.b = colour.b*g_b_coeff;

          // rescale to white level as saturation level
          colour.rgb = colour.rgb/(white_level-black_level);
          
          // apply gamma
          colour.rgb = pow(colour.rgb, vec3(1.0/gamma).rgb);

        }
    """

    fragmentShader_RAW = """
        #version 330 core
        
        in vec2 UV;
        uniform sampler2D backgroundTexture;
        uniform int channels; // channel representation
        uniform float white_level;
        uniform float black_level;
        uniform float g_r_coeff;
        uniform float g_b_coeff;
        uniform float max_value; // maximal value based on image precision
        uniform float max_type;  // maximal value based on image type (uint8, etc...)
        uniform float gamma;
        out vec3 colour;

        void main() {

           const int CH_RGGB = 4; // phase 0, bayer 2
           const int CH_GRBG = 5; // phase 1, bayer 3 (Boilers)
           const int CH_GBRG = 6; // phase 2, bayer 0
           const int CH_BGGR = 7; // phase 3, bayer 1 (Coconuts)

          vec4 bayer = texture(backgroundTexture, UV);
          // transform bayer data to RGB
          int r,gr,gb,b;
          switch (channels) {
            case 4:   r = 0; gr = 1; gb = 2; b = 3;  break; // CH_RGGB = 4 phase 0, bayer 2
            case 5:   r = 1; gr = 0; gb = 3; b = 2;  break; // CH_GRBG = 5 phase 1, bayer 3 (Boilers)
            case 6:   r = 2; gr = 3; gb = 0; b = 1;  break; // CH_GBRG = 6 phase 2, bayer 0
            case 7:   r = 3; gr = 2; gb = 1; b = 0;  break; // CH_BGGR = 7 phase 3, bayer 1 (Coconuts)
            default:        r = 0; gr = 1; gb = 2; b = 3;  break; // this should not happen
          }

          // first retreive black point to get the coefficients right ...
          // 5% of dynamics?
          
          // bayer 2 rgb
          colour.r   = bayer[r];
          colour.g = (bayer[gr]+bayer[gb])/2.0;
          colour.b = bayer[b];

          // black level
          colour.rgb = colour.rgb/max_value*max_type;
          colour.rgb = max((colour.rgb-vec3(black_level).rgb),0);
          
          // white balance
          colour.r = colour.r*g_r_coeff;
          colour.b = colour.b*g_b_coeff;

          // rescale to white level as saturation level
          colour.rgb = colour.rgb/(white_level-black_level);
          
          // apply gamma
          colour.rgb = pow(colour.rgb, vec3(1.0/gamma).rgb);
        }
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setAutoFillBackground(False)
        self.textureID = None
        self.tex_width, self.tex_height = 0, 0
        self.opengl_debug = False
        self.pMatrix  = np.identity(4, dtype=np.float32)
        self.mvMatrix = np.identity(4, dtype=np.float32)
        self.program_RGB = None
        self.program_RAW = None
        self.program = None
        self.vertexBuffer = None

    def set_shaders(self):
        if self.program_RGB is None:
            vs = shaders.compileShader(self.vertexShader, gl.GL_VERTEX_SHADER)
            fs = shaders.compileShader(self.fragmentShader_RGB, gl.GL_FRAGMENT_SHADER)
            try:
                self.program_RGB = shaders.compileProgram(vs, fs, validate=False)
                print("\n***** self.program_RGB = {} *****\n".format(self.program_RGB))
            except Exception as e:
                print('failed RGB shaders.compileProgram() {}'.format(e))
            shaders.glDeleteShader(vs)
            shaders.glDeleteShader(fs)

        if self.program_RAW is None:
            vs = shaders.compileShader(self.vertexShader, gl.GL_VERTEX_SHADER)
            fs = shaders.compileShader(self.fragmentShader_RAW, gl.GL_FRAGMENT_SHADER)
            try:
                self.program_RAW = shaders.compileProgram(vs, fs, validate=False)
                print("\n***** self.program_RAW = {} *****\n".format(self.program_RAW))
            except Exception as e:
                print('failed RAW shaders.compileProgram() {}'.format(e))
            shaders.glDeleteShader(vs)
            shaders.glDeleteShader(fs)

    def setVerticesBufferData(self):
        try:
            x0, x1, y0, y1 = self.image_centered_position()
            # print(" x0, x1, y0, y1 {} {} {} {}".format(x0, x1, y0, y1))
        except Exception as e:
            print(" Failed image_centered_position() {}".format(e))
            x0, x1, y0, y1 = 0, 100, 0, 100
            # set background vertices
        backgroundVertices = [
            x0, y1, 0.0,
            x0, y0, 0.0,
            x1, y1, 0.0,
            x1, y1, 0.0,
            x0, y0, 0.0,
            x1, y0, 0.0]
        vertexData = np.array(backgroundVertices, np.float32)

        if self.vertexBuffer is not None:
            self.vertexBuffer.destroy()
        self.vertexBuffer = QOpenGLBuffer()
        self.vertexBuffer.create()
        self.vertexBuffer.bind()
        self.vertexBuffer.allocate(vertexData, 4 * len(vertexData))

    def setBufferData(self):
        # set background UV
        backgroundUV = [
            0.0, 0.0,
            0.0, 1.0,
            1.0, 0.0,
            1.0, 0.0,
            0.0, 1.0,
            1.0, 1.0]
        uvData = np.array(backgroundUV, np.float32)

        self.uvBuffer = QOpenGLBuffer()
        self.uvBuffer.create()
        self.uvBuffer.bind()
        self.uvBuffer.allocate(uvData, 4 * len(uvData))

    def setTexture(self):
        texture_ok = super(GLImageViewerShaders, self).setTexture()
        self.setVerticesBufferData()
        return texture_ok

    def resizeGL(self, width, height):
        """Called upon window resizing: reinitialize the viewport.
        """
        # print(f"resizeGL {width}x{height}")
        if self.trace_calls:
            t = trace_method(self.tab)
        self._width = width*self.devicePixelRatio()
        self._height = height*self.devicePixelRatio()
        self.setVerticesBufferData()
        self.update()

    def initializeGL(self):
        """
        Initialize OpenGL, VBOs, upload data on the GPU, etc.
        """
        self.start_timing()

        time1 = get_time()
        self.set_shaders()
        self.add_time('set_shaders', time1)

        self.setVerticesBufferData()
        self.setBufferData()
        self.print_timing()

    def viewer_update(self):
        self.update()

    def paintGL(self):
        self.paintAll()

    def myPaintGL(self):
        """Paint the scene.
        """
        if self.textureID is None or not self.isValid():
            print("paintGL() not ready")
            return
        self.opengl_error()
        self.start_timing()

        gl.glClear(gl.GL_COLOR_BUFFER_BIT)

        if self._image.data.shape[2] == 4:
            self.program = self.program_RAW
        else:
            # TODO: check for other types: scalar ...
            self.program = self.program_RGB

        # Obtain uniforms and attributes
        self.aVert              = shaders.glGetAttribLocation(self.program, "vert")
        self.aUV                = shaders.glGetAttribLocation(self.program, "uV")
        self.uPMatrix           = shaders.glGetUniformLocation(self.program, 'pMatrix')
        self.uMVMatrix          = shaders.glGetUniformLocation(self.program, "mvMatrix")
        self.uBackgroundTexture = shaders.glGetUniformLocation(self.program, "backgroundTexture")
        self.channels_location    = shaders.glGetUniformLocation(self.program, "channels")
        self.black_level_location = shaders.glGetUniformLocation(self.program, "black_level")
        self.white_level_location = shaders.glGetUniformLocation(self.program, "white_level")
        self.g_r_coeff_location   = shaders.glGetUniformLocation(self.program, "g_r_coeff")
        self.g_b_coeff_location   = shaders.glGetUniformLocation(self.program, "g_b_coeff")
        self.max_value_location   = shaders.glGetUniformLocation(self.program, "max_value")
        self.max_type_location    = shaders.glGetUniformLocation(self.program, "max_type")
        self.gamma_location       = shaders.glGetUniformLocation(self.program, "gamma")

        # use shader program
        self.print_log("self.program = {}".format(self.program))
        shaders.glUseProgram(self.program)

        # set uniforms
        gl.glUniformMatrix4fv(self.uPMatrix, 1, gl.GL_FALSE, self.pMatrix)
        gl.glUniformMatrix4fv(self.uMVMatrix, 1, gl.GL_FALSE, self.mvMatrix)
        gl.glUniform1i(self.uBackgroundTexture, 0)

        gl.glUniform1i( self.channels_location, self._image.channels)

        # set color transformation parameters
        self.print_log("levels {} {}".format(self.filter_params.black_level.value,
                                             self.filter_params.white_level.value))
        gl.glUniform1f( self.black_level_location, self.filter_params.black_level.float)
        gl.glUniform1f( self.white_level_location, self.filter_params.white_level.float)

        # white balance coefficients
        gl.glUniform1f(self.g_r_coeff_location, self.filter_params.g_r.float)
        gl.glUniform1f(self.g_b_coeff_location, self.filter_params.g_b.float)

        # Should work for unsigned types for the moment
        gl.glUniform1f( self.max_value_location, (1 << self._image.precision)-1)
        gl.glUniform1f( self.max_type_location,  np.iinfo(self._image.data.dtype).max)

        gl.glUniform1f( self.gamma_location,       self.filter_params.gamma.float)

        # enable attribute arrays
        gl.glEnableVertexAttribArray(self.aVert)
        gl.glEnableVertexAttribArray(self.aUV)

        # set vertex and UV buffers
        # vert_buffers = VertexBuffers()
        # vert_buffers.vert_pos_buffer = vert_pos_buffer
        # vert_buffers.normal_buffer = normal_buffer
        # vert_buffers.tex_coord_buffer = tex_coord_buffer
        # vert_buffers.amount_of_vertices = int(len(index_array) / 3)

        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.vertexBuffer.bufferId())
        gl.glVertexAttribPointer(self.aVert, 3, gl.GL_FLOAT, gl.GL_FALSE, 0, None)
        gl.glBindBuffer(gl.GL_ARRAY_BUFFER, self.uvBuffer.bufferId())
        gl.glVertexAttribPointer(self.aUV, 2, gl.GL_FLOAT, gl.GL_FALSE, 0, None)

        # bind background texture
        # gl.glActiveTexture(gl.GL_TEXTURE0)
        gl.glBindTexture(gl.GL_TEXTURE_2D, self.textureID)
        gl.glEnable(gl.GL_TEXTURE_2D)

        # draw
        gl.glDrawArrays(gl.GL_TRIANGLES, 0, 6)

        # disable attribute arrays
        gl.glDisableVertexAttribArray(self.aVert)
        gl.glDisableVertexAttribArray(self.aUV)
        gl.glDisable(gl.GL_TEXTURE_2D)

        shaders.glUseProgram(0)

        self.print_timing(force=True)

    def updateTransforms(self) -> float:
        if self.trace_calls:
            t = trace_method(self.tab)
        if self.display_timing:
            start_time = get_time()
        self.makeCurrent()
        w = self._width
        h = self._height
        dx, dy = self.new_translation()
        # Deduce new scale from mouse vertical displacement
        scale = self.new_scale(-self.mouse_zoom_displ.y(), self.tex_height)
        # update the window size
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        translation_unit = min(w, h)/2
        gl.glScale(scale, scale, scale)
        gl.glTranslate(dx/translation_unit, dy/translation_unit, 0)
        # the window corner OpenGL coordinates are (-+1, -+1)
        gl.glOrtho(0, w, 0, h, -1, 1)
        self.pMatrix = np.array(gl.glGetFloatv(gl.GL_PROJECTION_MATRIX), dtype=np.float32).flatten()

        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadIdentity()
        self.mvMatrix = np.array(gl.glGetFloatv(gl.GL_MODELVIEW_MATRIX), dtype=np.float32).flatten()
        if self.display_timing:
            self.print_log('updateTransforms time {:0.1f} ms'.format((get_time()-start_time)*1000))
        return scale


if __name__ == '__main__':
    from qimview.image_readers import gb_image_reader
    # import numpy for generating random data points
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('input_image', help='input image')
    args = parser.parse_args()
    _params = vars(args)

    # define a Qt window with an OpenGL widget inside it
    # class TestWindow(QtGui.QMainWindow):
    class TestWindow(QtWidgets.QMainWindow):
        def __init__(self):
            super(TestWindow, self).__init__()
            self.widget = GLImageViewerShaders(self)
            self.show()
        def load(self):
            im = gb_image_reader.read(_params['input_image'])
            self.widget.set_image(im)
            self.widget.image_name = _params['input_image']
            # put the window at the screen position (100, 100)
            self.setGeometry(0, 0, self.widget._width, self.widget._height)
            self.setCentralWidget(self.widget)

    # create the Qt App and window
    app = QtWidgets.QApplication(sys.argv)
    window = TestWindow()
    window.load()
    window.show()
    app.exec_()


from qimview.utils.qt_imports import  QtWidgets
from typing import Optional

class FullScreenHelper:
    def __init__(self):
        # Variables used for fullscreen mode
        self._replacing_widget  : Optional[QtWidgets.QWidget] = None
        self._before_max_parent : Optional[QtWidgets.QWidget] = None

    def find_in_layout(self, layout: QtWidgets.QLayout, widget:QtWidgets.QWidget) -> Optional[QtWidgets.QLayout]:
        """ Search Recursivement in Layouts for the current widget
        Args:    layout (QtWidgets.QLayout): input layout for search
        Returns: layout containing the current widget or None if not found
        """
        if layout.indexOf(widget) != -1: return layout
        for i in range(layout.count()):
            item = layout.itemAt(i)
            if item.widget() == widget: return layout
            if (l := item.layout()) and (found:=self.find_in_layout(l, widget)): return l
        return None

    def toggle_fullscreen(self, widget:QtWidgets.QWidget) -> bool:
        if self._before_max_parent is None: 
            return self.enter_fullscreen(widget)
        else:
            return self.exit_fullscreen(widget)

    def enter_fullscreen(self, widget:QtWidgets.QWidget) -> bool:
        # if before_max_parent is not None, we are already in fullscreen mode
        if self._before_max_parent is not None: return False
        parent : Optional[QtWidgets.QWidget] = widget.parent()
        if parent is not None and (playout := parent.layout()) is not None:
            if self.find_in_layout(playout, widget):
                self._before_max_parent = parent
                self._replacing_widget = QtWidgets.QWidget(self._before_max_parent)
                playout.replaceWidget(widget, self._replacing_widget)
                # We need to go up from the parent widget to the main window to get its geometry
                # so that the fullscreen is display on the same monitor
                toplevel_parent : Optional[QtWidgets.QWidget] = widget.parentWidget()
                while toplevel_parent.parentWidget(): toplevel_parent = toplevel_parent.parentWidget()
                widget.setParent(None)
                if toplevel_parent: widget.setGeometry(toplevel_parent.geometry())
                widget.showFullScreen()
                return True
        return False

    def exit_fullscreen(self, widget:QtWidgets.QWidget) -> bool:
        # if before_max_parent is None, we are already not in fullscreen mode
        if self._before_max_parent is None or self._replacing_widget is None: return False
        parent = self._before_max_parent
        widget.setParent(parent)
        parent.layout().replaceWidget(self._replacing_widget, widget)
        self._replacing_widget   = None
        self._before_max_parent = None
        # self.resize(self.before_max_size)
        parent.update()
        widget.show()
        # Reset active main window 
        widget.activateWindow()
        widget.setFocus()
        return True




import sys, importlib
from pathlib import Path

def import_parents(current_file, level=1):
    global __package__
    file = Path(current_file).resolve()
    parent, top = file.parent, file.parents[level]
    
    sys.path.append(str(top))
    try:
        sys.path.remove(str(parent))
    except ValueError: # already removed
        pass

    __package__ = '.'.join(parent.parts[len(top.parts):])
    importlib.import_module(__package__) # won't be needed after that



from .qt_imports import *


class MVLabel(QLabel):
	'''
	This Class is a standard QLabel with the simple and double click mouse events
	created for MultiView class
	'''
	def __init__(self, text, parent = None):
		QLabel.__init__(self, text, parent)
		self.message = ""

	def mousePressEvent(self, event):
		self.last = "Click"

	def mouseReleaseEvent(self, event):
		print("mouseReleaseEvent")
		if self.last == "Click":
			QTimer.singleShot(QApplication.instance().doubleClickInterval(), self.performSingleClickAction)
		else:
			# Perform double click action.
			self.message = "Double Click"
			self.update()

	def mouseDoubleClickEvent(self, event):
		self.last = "Double Click"

	def performSingleClickAction(self):
		if self.last == "Click":
			self.message = "Click"
			self.update()



#
#
# started from https://cyrille.rossant.net/2d-graphics-rendering-tutorial-with-pyopengl/
#
# check also https://doc.qt.io/archives/4.6/opengl-overpainting.html
#

from qimview.utils.qt_imports   import QtWidgets, QOpenGLWidget, QtCore, QtGui

import OpenGL
OpenGL.ERROR_ON_COPY = True
import OpenGL.GL as gl
import OpenGL.GLU as glu
import traceback
import numpy as np
from typing import Optional, Tuple

from qimview.utils.viewer_image import ImageFormat, ViewerImage
from qimview.image_viewers.image_viewer import ImageViewer, trace_method


class GLImageViewerBase(ImageViewer, QOpenGLWidget, ):

    def __init__(self, parent : Optional[QtWidgets.QWidget] = None):
        ImageViewer.__init__(self, self)
        QOpenGLWidget.__init__(self, parent)
        self.setAutoFillBackground(False)

        _format = QtGui.QSurfaceFormat()
        print('profile is {}'.format(_format.profile()))
        print('version is {}'.format(_format.version()))
        # _format.setDepthBufferSize(24)
        # _format.setVersion(4,0)
        # _format.setProfile(PySide2.QtGui.QSurfaceFormat.CoreProfile)
        _format.setProfile(QtGui.QSurfaceFormat.CompatibilityProfile)
        self.setFormat(_format)

        # self.setFormat()
        self.textureID  = None
        self.tex_width, self.tex_height = 0, 0
        self.opengl_debug = True
        self.current_text = None
        self.cursor_imx_ratio = 0.5
        self.cursor_imy_ratio = 0.5
        self.trace_calls = False
        self.setMouseTracking(True)
        self.setFocusPolicy(QtCore.Qt.FocusPolicy.ClickFocus)

    # def __del__(self):
    #     if self.textureID is not None:
    #         gl.glDeleteTextures(np.array([self.textureID]))

    def set_image(self, image):
        if self.trace_calls:
            t = trace_method(self.tab)
        changed = super(GLImageViewerBase, self).set_image(image)

        if self._image is None:
            return
        img_width = self._image.data.shape[1]
        if img_width % 4 != 0:
            print("Image is resized to a multiple of 4 dimension in X")
            img_width = ((img_width >> 4) << 4)
            im = np.ascontiguousarray(self._image.data[:,:img_width, :])
            self._image = ViewerImage(im, precision = self._image.precision, downscale = 1,
                                        channels = self._image.channels)
            print(self._image.data.shape)

        if changed:  # and self.textureID is not None:
            print(f"self.textureID {self.textureID}")
            if self.setTexture():
                self.show()
                self.update()
            else:
                print("setTexture() return False")

    def synchronize_data(self, other_viewer):
        super(GLImageViewerBase, self).synchronize_data(other_viewer)
        other_viewer.cursor_imx_ratio = self.cursor_imx_ratio
        other_viewer.cursor_imy_ratio = self.cursor_imy_ratio

    def opengl_error(self, force=False):
        if self.opengl_debug or force:
            status = gl.glGetError()
            if status != gl.GL_NO_ERROR:
                print(self.tab[0]+'gl error %s' % status)

    def setTexture(self) -> bool:
        """
        :return: set opengl texture based on input numpy array image
        """
        # gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_GENERATE_MIPMAP_SGIS, gl.GL_TRUE)

        if self.trace_calls:
            t = trace_method(self.tab)
        self.start_timing()
        self.makeCurrent()

        # Replace texture only if required
        if self._image is None:
            print("self._image is None")
            return False

        img_height, img_width = self._image.data.shape[:2]

        # default type
        gl_types = {
            'int8'   : gl.GL_BYTE,
            'uint8'  : gl.GL_UNSIGNED_BYTE,
            'int16'  : gl.GL_SHORT,
            'uint16' : gl.GL_UNSIGNED_SHORT,
            'int32'  : gl.GL_INT,
            'uint32' : gl.GL_UNSIGNED_INT,
            'float32': gl.GL_FLOAT,
            'float64': gl.GL_DOUBLE
        }
        gl_type = gl_types[self._image.data.dtype.name]

        # It seems that the dimension in X should be even

        # Not sure what is the right parameter for internal format of 2D texture based
        # on the input data type uint8, uint16, ...
        # need to test with different DXR images
        internal_format = gl.GL_RGB
        if self._image.data.shape[2] == 3:
            if self._image.precision == 8:
                internal_format = gl.GL_RGB
            if self._image.precision == 10:
                internal_format = gl.GL_RGB10
            if self._image.precision == 12:
                internal_format = gl.GL_RGB12
        if self._image.data.shape[2] == 4:
            if self._image.precision == 8:
                internal_format = gl.GL_RGBA
            else:
                if self._image.precision <= 12:
                    internal_format = gl.GL_RGBA12
                else:
                    if self._image.precision <= 16:
                        internal_format = gl.GL_RGBA16
                    else:
                        internal_format = gl.GL_RGBA32F

        channels2format = {
            ImageFormat.CH_RGB  : gl.GL_RGB,
            ImageFormat.CH_BGR  : gl.GL_BGR,
            ImageFormat.CH_Y    : gl.GL_RED, # not sure about this one
            ImageFormat.CH_RGGB : gl.GL_RGBA, # we save 4 component data
            ImageFormat.CH_GRBG : gl.GL_RGBA,
            ImageFormat.CH_GBRG : gl.GL_RGBA,
            ImageFormat.CH_BGGR : gl.GL_RGBA
        }
        texture_pixel_format = channels2format[self._image.channels]

        if (self.tex_width,self.tex_height) != (img_width,img_height):
            try:
                if self.textureID is not None:
                    gl.glDeleteTextures(np.array([self.textureID]))
                self.textureID = gl.glGenTextures(1)
                gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 4)
                gl.glBindTexture(gl.GL_TEXTURE_2D, self.textureID)
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_BASE_LEVEL, 0)
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAX_LEVEL, 0)
                # gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAX_LEVEL, 10)
                # gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
                # gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST_MIPMAP_NEAREST)
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_NEAREST)
                gl.glTexImage2D(gl.GL_TEXTURE_2D, 0,
                                internal_format,
                                img_width, img_height,
                            0, texture_pixel_format, gl_type, self._image.data)
                # gl.glGenerateMipmap(gl.GL_TEXTURE_2D)
                self.tex_width, self.tex_height = img_width, img_height
            except Exception as e:
                print("setTexture failed shape={}: {}".format(self._image.data.shape, e))
                return False
        else:
            try:
                gl.glTexSubImage2D(gl.GL_TEXTURE_2D, 0, 0, 0, img_width, img_height,
                             texture_pixel_format, gl_type, self._image.data)
                # gl.glGenerateMipmap(gl.GL_TEXTURE_2D)
            except Exception as e:
                print("setTexture failed shape={}: {}".format(self._image.data.shape, e))
                return False

        self.print_timing(add_total=True)
        self.opengl_error()
        return True

    # @abstract_method
    def viewer_update(self):
        self.update()

    # To be defined in children
    def myPaintGL(self):  pass 

    def paintAll(self):
        if self.trace_calls:
            t = trace_method(self.tab)
        if self.textureID is None or not self.isValid() or not self.isVisible():
            print("paintGL()** not ready {} {}".format(self.textureID, self.isValid()))
            return
        # No need for makeCurrent() since it is called from PaintGL() only
        # self.makeCurrent()
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.beginNativePainting()
        im_pos = None
        scale  = None
        try:
            self.updateViewPort()
            scale = self.updateTransforms()
            self.myPaintGL()
            if self.show_cursor:
                im_pos = self.gl_draw_cursor()
        except Exception as e:
            self.print_log(" failed paintGL {}".format(e))
            traceback.print_exc()
        painter.endNativePainting()
        # status = gl.glGetError()

        draw_text = True
        if draw_text:
            # adapt scale depending on the ratio image / viewport
            scale *= self._width/self._image.data.shape[1]
            self.display_text(painter, self.display_message(im_pos, scale))

        # draw histogram
        if self.show_histogram:
            current_image = self._image.data
            rect = QtCore.QRect(0, 0, self.width(), self.height())
            histograms = self.compute_histogram_Cpp(current_image, show_timings=self.display_timing)
            self.display_histogram(histograms, 1,  painter, rect, show_timings=self.display_timing)

        painter.end()
        # Seems required here, at least on linux
        # self.update()

    def set_cursor_image_position(self, cursor_x, cursor_y):
        """
        Sets the image position from the cursor in proportion of the image dimension
        :return:
        """
        self.updateTransforms()
        ratio = self.screen().devicePixelRatio()
        self.print_log("ratio {}".format(ratio))
        pos_x = cursor_x * ratio
        pos_y = (self.height() - cursor_y) * ratio
        self.print_log("pos {} {}".format(pos_x, pos_y))
        x0, x1, y0, y1 = self.image_centered_position()

        gl_posX, gl_posY = self.get_mouse_gl_coordinates(pos_x, pos_y)
        self.cursor_imx_ratio = (gl_posX - x0) / (x1 - x0)
        self.cursor_imy_ratio = 1 - (gl_posY - y0) / (y1 - y0)
        self.print_log("cursor ratio {} {}".format(self.cursor_imx_ratio, self.cursor_imy_ratio))


    def gl_draw_cursor(self) -> Optional[Tuple[int, int]]:
        x0, x1, y0, y1 = self.image_centered_position()

        im_x = int(self.cursor_imx_ratio*self.tex_width)
        im_y = int(self.cursor_imy_ratio*self.tex_height)

        glpos_from_im_x = (im_x+0.5)*(x1-x0)/self.tex_width + x0
        glpos_from_im_y = (self.tex_height - (im_y+0.5))*(y1-y0)/self.tex_height+y0

        # get image coordinates
        length = 20 # /self.current_scale
        width = 4
        gl.glLineWidth(width)
        gl.glColor3f(0.0, 1.0, 1.0)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex3f(glpos_from_im_x-length, glpos_from_im_y, -0.001)
        gl.glVertex3f(glpos_from_im_x+length, glpos_from_im_y, -0.001)
        gl.glVertex3f(glpos_from_im_x, glpos_from_im_y-length, -0.001)
        gl.glVertex3f(glpos_from_im_x, glpos_from_im_y+length, -0.001)
        gl.glEnd()
        if im_x>=0 and im_x<self.tex_width and im_y>=0 and im_y<=self.tex_height:
            return (im_x, im_y) 
        return None


    def image_centered_position(self):
        w = self._width
        h = self._height
        self.print_log(f'self width height {self._width} {self._height} tex {self.tex_width} {self.tex_height}')
        if self.tex_width == 0 or self.tex_height == 0:
            return 0, w, 0, h
        # self.print_log(' {}x{}'.format(w, h))
        image_ratio = float(self.tex_width)/float(self.tex_height)
        if h*image_ratio < w:
            view_width  = int(h*image_ratio+0.5)
            view_height = h
            start_x = int((w - view_width) / 2 + 0.5)
            start_y = 0
        else:
            view_width  = w
            view_height = int(w/image_ratio+0.5)
            start_x = 0
            start_y = int((h - view_height) / 2 + 0.5)

        x0 = start_x
        x1 = start_x+view_width
        y0 = start_y
        y1 = start_y+view_height
        return x0, x1, y0, y1

    def updateViewPort(self):
        if self.trace_calls:
            t = trace_method(self.tab)
        self.start_timing()
        # keep image proportions
        w = self._width
        h = self._height
        # print(f" w, h {w, h}")
        try:
            gl.glViewport(0,0,w,h) # keep everything in viewport
        except Exception as e:
            self.print_log(" failed glViewport {}".format(e))
        self.print_timing(add_total=True)

    def updateTransforms(self) -> float:
        if self.trace_calls:
            t = trace_method(self.tab)
        self.start_timing()
        self.makeCurrent()
        w = self._width
        h = self._height
        dx, dy = self.new_translation()
        scale = self.new_scale(-self.mouse_zoom_displ.y(), self.tex_height)
        print(f"updateTransforms scale {scale}")
        try:
            # print("current context ", QtOpenGL.QGLContext.currentContext())
            # gl = QtOpenGL.QGLContext.currentContext().functions()
            # update the window size
            gl.glMatrixMode(gl.GL_PROJECTION)
            gl.glLoadIdentity()
            translation_unit = min(w, h)/2
            # self.print_log("scale {}".format(scale))
            gl.glScale(scale, scale, scale)
            gl.glTranslate(dx/translation_unit, dy/translation_unit, 0)
            # the window corner OpenGL coordinates are (-+1, -+1)
            gl.glOrtho(0, w, 0, h, -1, 1)
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glLoadIdentity()
        except Exception as e:
            self.print_log(" setting gl matrices failed {}".format(e))
        self.print_timing(add_total=True)
        self.opengl_error()
        return scale

    def resizeGL(self, width, height):
        """Called upon window resizing: reinitialize the viewport.
        """
        # print("ResizeGL")
        if self.trace_calls:
            t = trace_method(self.tab)
        # size give for opengl are in pixels, qt uses device independent size otherwise
        print(f"self.devicePixelRatio() {self.devicePixelRatio()}")
        self._width = width*self.devicePixelRatio()
        self._height = height*self.devicePixelRatio()
        # print("width height ratios {} {}".format(self._width/self.width(), self._height/self.height()))
        self.viewer_update()

    def get_mouse_gl_coordinates(self, x, y):
        modelview = gl.glGetDoublev(gl.GL_MODELVIEW_MATRIX)
        projection = gl.glGetDoublev(gl.GL_PROJECTION_MATRIX)
        viewport = gl.glGetIntegerv(gl.GL_VIEWPORT)
        posX, posY, posZ = glu.gluUnProject(x, y, 0, modelview, projection, viewport)
        return posX, posY

    def mousePressEvent(self, event):
        self._mouse_events.mouse_press_event(event)

    def mouseMoveEvent(self, event):
        if self.show_cursor:
            self.set_cursor_image_position(event.x(), event.y())
        self._mouse_events.mouse_move_event(event)

    def mouseReleaseEvent(self, event):
        self._mouse_events.mouse_release_event(event)

    def mouseDoubleClickEvent(self, event):
        self._mouse_events.mouse_double_click_event(event)

    def wheelEvent(self, event):
        self._mouse_events.mouse_wheel_event(event)

    def keyPressEvent(self, event):
        # TODO: Fix the correct parameters for selecting image zoom/pan
        x0, x1, y0, y1 = self.image_centered_position()
        print(f"image centered position {x1-x0} x {y1-y0}")
        self.key_press_event(event, wsize=QtCore.QSize(x1-x0, y1-y0))

    def resizeEvent(self, event):
        """Called upon window resizing: reinitialize the viewport.
        """
        if self.trace_calls:
            t = trace_method(self.tab)
        self.print_log(f"resize {event.size()}  self {self.width()} {self.height()}")
        self.evt_width = event.size().width()
        self.evt_height = event.size().height()
        QOpenGLWidget.resizeEvent(self, event)
        self.print_log(f"resize {event.size()}  self {self.width()} {self.height()}")





from qimview.utils.qt_imports import *
from qimview.parameters.numeric_parameter_gui import NumericParameterGui

class ImageFilterParametersGui:
    def __init__(self, parameters, name=""):
        """
        :param parameters: instance of ImageFilterParameters
        """
        self.params    = parameters
        self.bl_gui    = None
        self.wl_gui    = None
        self.gamma_gui = None
        self.g_r_gui   = None
        self.g_b_gui   = None
        self.saturation_gui    = None
        self.imdiff_factor_gui = None
        self.event_recorder    = None
        self.name              = name

    def set_event_recorder(self, evtrec):
        self.event_recorder = evtrec

    def add_blackpoint(self, layout, callback):
        self.bl_gui = NumericParameterGui("Black", self.params.black_level, callback, layout, self.name)
        self.bl_gui.set_event_recorder(self.event_recorder)

    def add_whitepoint(self, layout, callback):
        self.wl_gui = NumericParameterGui("White", self.params.white_level, callback, layout, self.name)
        self.wl_gui.set_event_recorder(self.event_recorder)

    def add_gamma(self, layout, callback):
        self.gamma_gui = NumericParameterGui("Gamma", self.params.gamma, callback, layout, self.name)
        self.gamma_gui.set_event_recorder(self.event_recorder)

    def add_g_r(self, layout, callback):
        self.g_r_gui = NumericParameterGui("G/R", self.params.g_r, callback, layout, self.name)
        self.g_r_gui.set_event_recorder(self.event_recorder)

    def add_g_b(self, layout, callback):
        self.g_b_gui = NumericParameterGui("G/B", self.params.g_b, callback, layout, self.name)
        self.g_b_gui.set_event_recorder(self.event_recorder)

    def add_saturation(self, layout, callback):
        self.saturation_gui = NumericParameterGui("Saturation", self.params.saturation, callback, layout, self.name)
        self.saturation_gui.set_event_recorder(self.event_recorder)

    def add_imdiff_factor(self, layout, callback):
        self.imdiff_factor_gui = NumericParameterGui("Image diff factor", self.params.imdiff_factor, callback, layout, self.name)
        self.imdiff_factor_gui.set_event_recorder(self.event_recorder)

    def register_event_player(self, event_player):
        for v in vars(self):
            if 'gui' in v and self.__dict__[v]:
                self.__dict__[v].register_event_player(event_player)

    def reset_all(self):
        for v in vars(self):
            if 'gui' in v:
                self.__dict__[v].reset()




from ..utils.qt_imports import QtCore, QtWidgets
from .qtdump import *

class EventPlayer:

    def __init__(self):
        self.widgets = dict()

    def register_widget(self, name, widget):
        self.widgets[name] = widget

    def play_events(self, event_list=None, start_delay=100):
        if len(event_list) > 0:
            # Send events using Qtimer
            print("EventPlayer: start")
            QtCore.QTimer.singleShot(start_delay, lambda: self.send_event( event_list, 0))


    def send_event(self, event_list, index):
        # now create mouse event
        current_event = event_list[index]
        event_info    = current_event['info']
        widget_name   = current_event.get('widget_name')
        # print(f"EventPlayer {index}: {widget_name} {event_info}", end=', ')
        if widget_name in self.widgets:
            widget = self.widgets[widget_name]
            if isinstance(event_info['type'],str):
                if event_info['type'] == 'save_screen':
                    widget_hash = QtDump.get_screen_hash(widget)
                    recorded_hash = event_info['hash']
                    # print(f"hash now {widget_hash} recorded {recorded_hash}")
                    if widget_hash != recorded_hash:
                        print('Hash comparison failure !!!')
                    else:
                        print('Hash comparison success !!!')
            else:
                # --- Deal with QEvent
                qt_event = QtDump.dict2event(event_info)
                if qt_event.type() in RESIZE_EVENTS:
                    # for resize, we need to actually do the resize of the control widget instead of calling the resize event
                    # QtWidgets.QApplication.postEvent(widget.parent(), new_event)
                    # print(" resize", end=', ')
                    widget.resize(qt_event.size())
                else:
                    # print(" post event", end=', ')
                    QtWidgets.QApplication.postEvent(widget, qt_event)
        else:
            print(f"widget '{widget_name}' not found", end=', ')
        index = index + 1
        if index < len(event_list):
            # print("")
            delay = event_list[index]['time']-event_list[index-1]['time']
            QtCore.QTimer.singleShot(delay*1000,  lambda: self.send_event(event_list, index))
        else:
            print(" end")





class NumericParameter:
    """
    Numeric Parameter is specifically designed for sliders:
    it is a numeric value controlled by a limited number of possible integer values
    """
    def __init__(self, value=0, default_value=0, _range=[0,100], float_scale = 100):
        self.value = value
        self.default_value = default_value
        self.range = _range
        self.float_scale = float_scale

    @property
    def int(self):
        return self.value

    @int.setter
    def int(self, v):
        self.value = v

    @property
    def float(self):
        return float(self.value/self.float_scale)

    @float.setter
    def float(self, f):
        self.value = int(f*self.float_scale+0.5)

    def __repr__(self):
        return f"<NumericParameter val:{self.value} def:{self.default_value} rg:{self.range} sc:{self.float_scale}>"

    def __str__(self):
        return f"{self.float}"

    def copy_from(self, p):
        for v in vars(self):
            self.__dict__[v] = p.__dict__[v]






try:
  import rawpy
except:
  has_rawpy = False
else:
  has_rawpy = True
import math
import numpy as np
from qimview.utils.viewer_image import ViewerImage, ImageFormat
from io import BytesIO

def read_libraw(image_filename, image_buffer, read_size='full', use_RGB=True, verbose=False):
    if image_buffer:
        raw = rawpy.imread(BytesIO(image_buffer))
    else:
        raw = rawpy.imread(image_filename)
    height, width = raw.raw_image.shape
    print(f"height, width {(height, width)}")
    # Transform Bayer image
    im1 = np.empty((height >> 1, width >> 1, 4), dtype=raw.raw_image.dtype)
    bayer_desc = raw.color_desc.decode('utf-8')
    bayer_pattern = raw.raw_pattern
    # not perfect, we may switch Gr and Gb
    bayer_desc = bayer_desc.replace('G', 'g', 1)
    # Detect the green channels?
    channel_pos = {'R': 0, 'g': 1, 'G': 2, 'B': 3}
    for i in range(2):
        for j in range(2):
            print("str(bayer_desc[bayer_pattern[i,j]]) {}".format(
                str(bayer_desc[bayer_pattern[i, j]])))
            pos = channel_pos[str(bayer_desc[bayer_pattern[i, j]])]
            im1[:, :, pos] = raw.raw_image[i::2, j::2]
    prec = math.ceil(math.log2(raw.white_level+1))
    viewer_image = ViewerImage(im1, precision=prec, downscale=1, channels=ImageFormat.CH_RGGB)
    print("viewer_image channels {}".format(viewer_image.channels))
    return viewer_image


def libraw_supported_formats():
    # Need to find the complete list of supported formats
    return [".ARW", ".GPR", ".DNG" ]





from qimview.utils.utils import get_time
# from qimview.utils.qt_imports import *
from .basecache import BaseCache
import os
from os import path as osp
import psutil
from typing import Optional, Tuple

class FileCache(BaseCache[str,bytes,float]):
    """
        Save output bytes from read() function into a cache indexed by the filename
        inherits from BaseCache, with
            id as string: input filename
            bytes: output from binary read()
            mtime: modification time as float from osp.getmtime(filename)
        If a file is in the cache but its modification time on disk is more recent,
        we can enable an automatic reload

    Args:
        BaseCache (_type_): _description_
    """    
    def __init__(self):
        BaseCache.__init__(self, "FileCache")
        total_memory        : float = psutil.virtual_memory().total / self.cache_unit
        # let use 5% of total memory by default
        self.max_cache_size : int   = int(total_memory * 0.05)
        self.last_progress = 0
        self.verbose : bool = False

    def has_file(self, filename):
        # is it too slow
        filename = os.path.abspath(filename)
        return filename in self.cache_list

    def get_file(self, filename: str, check_size: bool = True) -> Tuple[Optional[bytes], bool]:
        """_summary_

        Args:
            filename (str): file to read, is supposed to exist on disk, if not an exception will be raised
                by Python standard methods
            check_size (bool, optional): _description_. Defaults to True.

        Returns:
            Tuple[Optional[bytes], bool]: first elt is the data read 
                second is a boolean saying if it comes or not from the cache 
        """
        # print(f'get_file {filename}')
        start = get_time()
        # Get absolute normalized path
        filename = os.path.abspath(filename)
        # Get file last modification time
        mtime = os.path.getmtime(filename)
        # print(f"image cache get_image({filename})")
        file_data = self.search(filename)
        if file_data is not None:
            if file_data[2]>=mtime:
                # print(f'get_file {filename} found end')
                return file_data[1], True
            else:
                # Remove element from cache?
                print("Removing outdated cache data from file {filename}")
                self.cache.remove(file_data)
                self.cache_list.remove(filename)
        try:
            # Read file as binary data
            self._print_log(" FileCache::get_file() before read() {0:0.3f} sec.".format(get_time() - start))
            with open(filename, 'rb') as f:
                file_data = f.read()
            self._print_log(" FileCache::get_file() after read() {0:0.3f} sec.".format(get_time() - start))
            self.append(filename, file_data, extra=mtime, check_size=check_size)
            self._print_log("  FileCache::get_file() after append took {0:0.3f} sec.".format(get_time() - start))
        except Exception as e:
            print("Failed to load image {0}: {1}".format(filename, e))
            return None, False
        else:
            # print(f'get_file {filename} read end')
            return file_data, False

    def thread_add_files(self, filenames, progress_callback = None):
        """
        :param filename:
        :param show_timing:
        :return: pair image_data, boolean (True is coming from cache)
        """
        nb_files = len(filenames)
        nb_new_files = 0
        for n, f in enumerate(filenames):
            if nb_new_files>80:
                break
            if f is not None and not self.has_file(f):
                data, flag = self.get_file(f, check_size=False)
                # slow down to simulate network drive
                # sleep(.500)
                if not flag:
                    nb_new_files += 1
            if progress_callback is not None:
                progress_callback.emit(int(n*100/nb_files+0.5))
            # if n%10==0: 
            #     self.check_size_limit()

    def file_added(self, filename):
        pass
        # print(f" *** Added image {os.path.basename(filename)} to cache finished")

    def add_result(self, r):
        # print(f"adding {r} to results ")
        pass

    def show_progress(self, val):
        if val!=self.last_progress:
            print(f"add files done {val} %")
            self.check_size_limit()
            self.last_progress = val

    def on_finished(self):
        self.check_size_limit()

    def add_files(self, filenames):
        self.thread_pool.clear()
        start = get_time()
        self.add_results = []
        # print(f" start worker with image {f}")
        # This part may be causing issues
        use_threads = True
        if use_threads:
            self.thread_pool.set_worker(self.thread_add_files, filenames)
            self.thread_pool.set_worker_callbacks(finished_cb=self.on_finished)
            self.thread_pool.start_worker()
        else:
            self.thread_add_files(filenames, progress_cb=self.show_progress)
        self._print_log(f" FileCache.add_files() {self.add_results} took {int((get_time()-start)*1000+0.5)} ms;")




from qimview.utils.qt_imports import QtGui, QtCore, QtWidgets
from typing import TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from .multi_view import MultiView
QtKeys  = QtCore.Qt.Key
QtMouse = QtCore.Qt.MouseButton


class MultiViewKeyEvents:
    """ Implement events for MultiView
    """
    def __init__(self, multiview: 'MultiView'):
        self._multiview : 'MultiView' = multiview
        # Set key events callbacks
        self.keys_callback = {
                'F1'             : self.helpDialog,
                'F5'             : self.reloadImages,
                'F10'            : self.toggleFullScreen,
                'Esc'            : self.exitFullScreen,
                'Up'             : self.upCallBack,
                'Down'           : self.downCallBack,
                'Left'           : self.selectPreviousImage,
                'Right'          : self.selectNextImage,
                'G'              : self.changeGridLayout,
        }
        # Use regular expression with 1 group that will be the argument of the callback
        # currently limited to function that take only one integer parameter to return a callback
        self.keys_callback_re = {
                'Alt\+Keypad\+([0-9])' : self.setActiveViewerImage,
                'Alt\+Shift\+([0-9])'  : self.setActiveViewerImage,
                'Ctrl\+Keypad\+([0-9])': self.setReferenceImage,
                'Ctrl\+Shift\+([0-9])' : self.setReferenceImage,
                'Keypad\+([1-9])'      : self.setNumberOfViewers,
                'Shift\+([1-9])'       : self.setNumberOfViewers,
        }

        self._help_links : str = "[wiki help: Multi-image Viewer](https://github.com/qimview/qimview/wiki/4.-Multi%E2%80%90image-viewer)  \n"

    @staticmethod
    def get_key_seq(event : QtGui.QKeyEvent) -> str:
        """ Return a key sequence from a keyboard event
        """
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        QtMod = QtCore.Qt.KeyboardModifier
        # Python > 3.7 key order is maintained
        mod_str = { 
            QtMod.ControlModifier : 'Ctrl',
            QtMod.AltModifier     : 'Alt',
            QtMod.MetaModifier    : 'Meta',
            QtMod.ShiftModifier   : 'Shift',
            QtMod.KeypadModifier  : 'Keypad',
        }
        # Compact line to create the contatenated string like 'Ctrl+Alt+'
        mod_seq_str = "".join([ f"{mod_str[m]}+" for  m in mod_str.keys() if modifiers & m])
        key_str = QtGui.QKeySequence(event.key()).toString()
        print(f'{mod_seq_str + key_str} event.text() = {event.text()}')
        return mod_seq_str + key_str
        # return QtGui.QKeySequence(mod_seq_str + key_str).toString()

    def _get_markdown_help(self) -> str:
        res = ''
        res += '|Keys  |Action  |  \n'
        res += '|:-----|:------:|  \n'
        # TODO create html table
        for k,v in self.keys_callback.items():
            res += f'|{k}|{v.__doc__}|  \n'
        for k,v in self.keys_callback_re.items():
            res += f'|{k}|{v.__doc__}|  \n'
        res += '  \n'
        return res
    
    def markdown_help(self) -> str:
        return self._get_markdown_help()

    def help_links(self) -> str:
        return self._help_links

    def helpDialog(self) -> bool :
        """ Open help dialog with links to wiki pages """
        import qimview
        mb = QtWidgets.QMessageBox(self._multiview)
        mb.setWindowTitle(f"qimview {qimview.__version__}: MultiView help")
        mb.setTextFormat(QtCore.Qt.TextFormat.MarkdownText)
        mb.setText(
                f"# qimview {qimview.__version__}  \n" +
                self._get_markdown_help() +
                "### Links \n" +
                self._help_links +
                "[github: qimview](https://github.com/qimview/qimview/wiki)  \n" +
                "[wiki help: Image Viewer](https://github.com/qimview/qimview/wiki/3.-Image-Viewers)  \n"
                )
        mb.exec()
        return True

    def reloadImages(self) -> bool:
        """ Reload images from disk """
        self._multiview.update_image(reload=True)
        return True
    
    def toggleFullScreen(self) -> bool:
        """ Toggle fullscreen """
        return self._multiview._fullscreen.toggle_fullscreen(self._multiview)
    
    def exitFullScreen(self) -> bool:
        """ Exit fullscreen """
        return self._multiview._fullscreen.exit_fullscreen (self._multiview)

    def upCallBack(self) -> bool:
        """ Call user-defined callback """
        if self._multiview.key_up_callback is not None:
            self._multiview.key_up_callback()
            return True
        return False 
    
    def downCallBack(self) -> bool:
        """ Call user-defined callback """
        if self._multiview.key_down_callback is not None:
            self._multiview.key_down_callback()
            return True
        return False 
    
    def setNumberOfViewers(self, n:int) -> Callable:
        """ Set the number N of displayed viewers """
        def func() -> bool:
            mv = self._multiview
            mv.set_number_of_viewers(n)
            if mv._active_viewer not in mv.image_viewers:
                mv._active_viewer = mv.image_viewers[n-1]
            mv.viewer_grid_layout.update()
            mv.update_image(mv._active_viewer.image_name)
            mv.setFocus()
            return True
        return func

    def setActiveViewerImage(self, n:int) -> Callable:
        """ Set the image N displayed by the active viewer """
        def func() -> bool:
            mv = self._multiview
            if n>=len(mv.image_list): return False
            if mv.image_list[n] is None: return False
            if mv.output_label_current_image != mv.image_list[n]:
                mv.update_image(mv.image_list[n])
                mv.setFocus()
                return True
            return False
        return func

    def setReferenceImage(self, n:int) -> Callable:
        """ Set the reference image N """
        def func() -> bool:
            mv = self._multiview
            if mv.image_list[n] is None: return False
            if mv.output_label_current_image != mv.image_list[n]:
                mv.set_reference_label(mv.image_list[n], update_viewers=True)
                mv.update_image()
                return True
            return False
        return func

    def selectPreviousImage(self) -> bool:
        """ Display previous image on active viewer """
        try:
            mv = self._multiview
            current_pos = mv.image_list.index(mv.output_label_current_image)
            nb_images = len(mv.image_list)
            mv.update_image(mv.image_list[(current_pos+nb_images-1)%nb_images])
            return True
        except ValueError:
            return False

    def selectNextImage(self) -> bool:
        """ Display next image on active viewer """
        try:
            mv = self._multiview
            current_pos = mv.image_list.index(mv.output_label_current_image)
            nb_images = len(mv.image_list)
            mv.update_image(mv.image_list[(current_pos+1)%nb_images])
            return True
        except ValueError:
            return False

    def changeGridLayout(self) -> bool:
        """ Change grid layout """
        mv = self._multiview
        mv.max_columns = int ((mv.max_columns + 1) % mv.nb_viewers_used + 1)
        mv.set_number_of_viewers(mv.nb_viewers_used, max_columns=mv.max_columns)
        mv.update_image(reload=True)
        mv.setFocus()
        return True

    def key_press_event(self, event : QtGui.QKeyEvent):
        if type(event) == QtGui.QKeyEvent:
            key_seq : str = MultiViewKeyEvents.get_key_seq(event)
            print(f"key_seq {key_seq}")
            if key_seq in self.keys_callback:
                event.setAccepted(self.keys_callback[key_seq]())
            else:
                import re
                # try with regular expressions
                for str in self.keys_callback_re:
                    e = re.compile(str)
                    m = e.match(key_seq)
                    if m and len(m.groups()) == 1:
                        val = int(m.group(1))
                        event.setAccepted(self.keys_callback_re[str](val)())
                        return
                event.ignore()
        else:
            event.ignore()



# Import from the less dependent to the most dependent module
from .image_filter_parameters     import ImageFilterParameters
from .image_filter_parameters_gui import ImageFilterParametersGui
from .qt_image_viewer             import QTImageViewer
from .gl_image_viewer             import GLImageViewer
from .gl_image_viewer_shaders     import GLImageViewerShaders
from .multi_view                  import MultiView, ViewerType

__all__ = [
    'ImageFilterParameters',
    'ImageFilterParametersGui',
    'QTImageViewer',
    'GLImageViewer',
    'GLImageViewerShaders',
    'ViewerType',
    'MultiView',
]



from qimview.utils.qt_imports import QtGui, QtCore, QtWidgets
from typing import TYPE_CHECKING, List, Tuple
from qimview.utils.tab_dialog import TabDialog
if TYPE_CHECKING:
    from .image_viewer import ImageViewer
QtKeys  = QtCore.Qt.Key
QtMouse = QtCore.Qt.MouseButton


class ImageViewerKeyEvents:
    """ Implement events for ImageViewer
    """
    def __init__(self, viewer: 'ImageViewer'):
        self._viewer : 'ImageViewer' = viewer
        # Set arbitrary size for initialization
        self._wsize : QtCore.QSize = QtCore.QSize(100,100)

        # Set key events callbacks
        self.keys_callback = {
                'A'     : self.toggleAntialiasing,
                'C'     : self.toggleCursor,
                'D'     : self.toggleDifferences,
                'H'     : self.toggleHistogram,
                'I'     : self.toggleIntensityLine,
                'O'     : self.toggleOverlay,
                'S'     : self.toggleStats,
                'F1'    : self.helpDialog,
                'F11'   : self.toggleFullScreen,
                'Esc'   : self.exitFullScreen,
                'Alt+A' : self.zoomUpperLeft,
                'Alt+B' : self.zoomUpperRight,
                'Alt+C' : self.zoomLowerLeft,
                'Alt+D' : self.zoomLowerRight,
                'Alt+F' : self.unZoom,
                'Ctrl+B': self.copy2Clipboard,
                'Ctrl+S': self.saveImage,
        }

        self._help_tabs  : List[Tuple[str,str]] = []
        self._help_links : str = ''

    def _get_markdown_help(self) -> str:
        res = ''
        res += '|Keys  |Action  |  \n'
        res += '|:-----|:------:|  \n'
        # TODO create html table
        for k,v in self.keys_callback.items():
            res += f'|{k}|{v.__doc__}|  \n'
        res += '  \n'
        return res
    
    def add_help_tab(self, title: str, text: str) -> None:
        """ Additional help to display
        Args:  help (str): help string in markdown format
        """
        self._help_tabs.append((title, text))

    def add_help_links(self, help_links: str) -> None:
        """ Additional help links to display
        Args:  help_links (str): help string in markdown format
        """
        self._help_links = help_links

    @staticmethod
    def get_key_seq(event : QtGui.QKeyEvent) -> QtGui.QKeySequence:
        """ Return a key sequence from a keyboard event
        """
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        QtMod = QtCore.Qt.KeyboardModifier
        # Python > 3.7 key order is maintained
        mod_str = { 
            QtMod.ShiftModifier   : 'Shift',
            QtMod.ControlModifier : 'Ctrl',
            QtMod.AltModifier     : 'Alt',
            QtMod.MetaModifier    : 'Meta',
        }
        # Compact line to create the contatenated string like 'Ctrl+Alt+'
        mod_seq_str = "".join([ f"{mod_str[m]}+" for  m in mod_str.keys() if modifiers & m])
        key_str = QtGui.QKeySequence(event.key()).toString()
        return QtGui.QKeySequence(mod_seq_str + key_str)
    
    def helpDialog(self) -> bool :
        """ Open help dialog with links to wiki pages """
        import qimview
        help_win = TabDialog(self._viewer.widget, f"qimview {qimview.__version__}: ImageViewer help")
        # mb.setTextFormat(QtCore.Qt.TextFormat.RichText)
        help_win.add_markdown_tab("ImageViewer keys",
                                  self._get_markdown_help()
                                  )
        for (title,text) in self._help_tabs:
            help_win.add_markdown_tab(title, text)

        help_win.add_markdown_tab("Links",
                "### Links \n" +
                "[github: qimview](https://github.com/qimview/qimview/wiki)  \n" +
                "[wiki help: Image Viewer](https://github.com/qimview/qimview/wiki/3.-Image-Viewers)  \n" +
                self._help_links
                )
        help_win.show()
        return True
    
    def toggleFullScreen(self) -> bool:
        """ toggle fullscreen mode """
        return self._viewer._fullscreen.toggle_fullscreen(self._viewer._widget)

    def exitFullScreen(self) -> bool: 
        """ exit fullscreen mode """
        return self._viewer._fullscreen.exit_fullscreen(self._viewer._widget)
    
    def updateAndAccept(self) -> bool:
        self._viewer.viewer_update()
        self._viewer.synchronize()
        return True
    
    def zoomUpperLeft(self) -> bool:
        """ Display upper left quarter of the image """
        self._viewer.current_dx = int(self._wsize.width()/4+0.5)
        self._viewer.current_dy = -int(self._wsize.height()/4+0.5)
        self._viewer.current_scale = 2
        return self.updateAndAccept()

    def zoomLowerLeft(self)    -> bool:
        """ Display lower left quarter of the image """
        self._viewer.current_dx = int(self._wsize.width()/4+0.5)
        self._viewer.current_dy = int(self._wsize.height()/4+0.5)
        self._viewer.current_scale = 2
        return self.updateAndAccept()

    def zoomUpperRight(self)    -> bool:
        """ Display upper right quarter of the image """
        self._viewer.current_dx = -int(self._wsize.width()/4+0.5)
        self._viewer.current_dy = -int(self._wsize.height()/4+0.5)
        self._viewer.current_scale = 2
        return self.updateAndAccept()

    def zoomLowerRight(self)    -> bool:
        """ Display lower right quarter of the image """
        self._viewer.current_dx = -int(self._wsize.width()/4+0.5)
        self._viewer.current_dy =  int(self._wsize.height()/4+0.5)
        self._viewer.current_scale = 2
        return self.updateAndAccept()

    def unZoom(self)    -> bool:
        """ Display full image """
        self._viewer.current_dx = 0
        self._viewer.current_dy = 0
        self._viewer.current_scale = 1
        return self.updateAndAccept()
    
    def toggleAntialiasing(self)->bool: 
        """ Toggle anti-aliasing """
        self._viewer.antialiasing   = not self._viewer.antialiasing
        return self.updateAndAccept()
    
    def toggleHistogram(self) ->bool:
        """ Toggle histogram display """
        self._viewer.show_histogram = not self._viewer.show_histogram
        return self.updateAndAccept()
    
    def toggleOverlay(self) ->bool:
        """ Toggle overlay display """
        self._viewer.show_overlay   = not self._viewer.show_overlay
        return self.updateAndAccept()
    
    def toggleCursor(self) ->bool:
        """ Toggle cursor display """
        self._viewer.show_cursor    = not self._viewer.show_cursor
        return self.updateAndAccept()
    
    def toggleStats(self) ->bool:
        """ Toggle stats display """
        self._viewer.show_stats     = not self._viewer.show_stats
        return self.updateAndAccept()

    def toggleDifferences(self) ->bool: 
        """ Toggle image difference with reference """
        self._viewer.show_image_differences = not self._viewer.show_image_differences
        return self.updateAndAccept()
    
    def toggleIntensityLine(self) ->bool: 
        """ Toggle horizontal intensity line display """
        self._viewer.show_intensity_line = not self._viewer.show_intensity_line
        return self.updateAndAccept()
    
    def copy2Clipboard(self)->bool:
        """ Copy current image to clipboard """
        clipboard = QtWidgets.QApplication.clipboard()
        self._viewer.set_clipboard(clipboard, True)
        self._viewer.widget.repaint()
        self._viewer.set_clipboard(None, False)
        print("Image saved to clipboard")
        return True

    def saveImage(self)->bool:
        """ Save current image to disk and clipboard """
        clipboard = QtWidgets.QApplication.clipboard()
        self._viewer.set_clipboard(clipboard, True)
        self._viewer.widget.repaint()
        self._viewer.set_clipboard(None, False)
        # Ask for input file
        filename = QtWidgets.QFileDialog.getSaveFileName(self._viewer._widget, "ImageViewer: Save current image", filter="Images (*.png *.xpm *.jpg)")
        print(f"filename {filename}")
        im = clipboard.image()
        if im.isNull():
            print("Image not found")
            return False
        else:
            return im.save(filename[0])

    def key_press_event(self, event : QtGui.QKeyEvent, wsize : QtCore.QSize):
        self._wsize = wsize
        # print(f"ImageViewerEvents: key_press_event {event.key()}")
        if type(event) == QtGui.QKeyEvent:

            key_seq : str = ImageViewerKeyEvents.get_key_seq(event).toString()
            # print(f"key sequence = {key_seq}")
            if key_seq in self.keys_callback:
                event.setAccepted(self.keys_callback[key_seq]())
            else:
                event.ignore()
        else:
            event.ignore()




from .qt_imports import *

class MenuSelection:
    def __init__(self, _title, _menu, _dict, _default, _callback=None):
        """[summary]

        Args:
            _title ([type]): [description]
            _menu ([type]): [description]
            _dict ([type]): [description]
            _default ([type]): [description]
            _callback ([type], optional): [description]. Defaults to None.
        """        
        self._options = _dict
        self._default = _default
        self._current = _default
        self._menu    = _menu
        self._menu.addSection(_title)
        self._group = QtGui.QActionGroup(_menu)
        self._actions = {}
        self._callback = _callback
        for option in self._options:
            action = QtGui.QAction(option,  self._menu, checkable=True)
            action.setActionGroup(self._group)
            self._menu.addAction(action)
            self._actions[option] = action
        self._group.triggered.connect(self.update_selection)
        self._actions[self._current].setChecked(True)

    def update_selection(self):
        print(f"MenuSelection update_selection()")
        # find selected option
        for option in self._options:
            if self._actions[option].isChecked():
                self._current = option
                print(f" selection is {option}")
        if self._callback is not None:
            self._callback()
    
    def get_selection(self):
        return self._current
    
    def get_selection_value(self):
        return self._options[self._current]
    
    def set_selection(self, selection):
        self._current = selection
        self._actions[self._current].setChecked(True)

    def setEnabled(self, enabled):
        for option in self._options:
            self._actions[option].setEnabled(enabled)





# required to install codecs to work well on windows, I installed K-Lite standard codecs

from ..utils.qt_imports import QtGui, QtCore, QtWidgets, QtMultimedia, QtMultimediaWidgets
from ..parameters.numeric_parameter import  NumericParameter
from ..parameters.numeric_parameter_gui import NumericParameterGui

import sys

class myVideoWidget(QtMultimediaWidgets.QVideoWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.current_scale = 1

    def new_scale(self, mouse_zy, height):
        return max(1, self.current_scale * (1 + mouse_zy * 5.0 / height))
        # return max(1, self.current_scale  + mouse_zy * 5.0 / height)

    def wheelEvent(self, event):
        if hasattr(event, 'delta'):
            delta = event.delta()
        else:
            delta = event.angleDelta().y()
        coeff = delta/5
        rect = self.geometry()
        prev_scale = self.current_scale
        self.current_scale = self.new_scale(coeff, rect.height())
        print(f" geometry {rect.x()} {rect.y()} {rect.width()} {rect.height()}")
        print(f" current_scale {self.current_scale}")
        rect.setWidth (rect.width() *(self.current_scale/prev_scale))
        rect.setHeight(rect.height()*(self.current_scale/prev_scale))
        # self.setGeometry(rect)
        self.setFixedSize(rect.width(), rect.height())

    def resizeEvent(self, event):
        """Called upon window resizing: reinitialize the viewport.
        """
        # print("resize {} {}  self {} {}".format(event.size().width(), event.size().height(),
        #       self.width(), self.height()))
        # event.ignore()

class VideoPlayer(QtWidgets.QWidget):
    def __init__(self, parent, open_button=False):
        super().__init__(parent)

        self.setWindowTitle("QMediaPlayer")
        # self.setGeometry(350, 100, 700, 500)
        self.setWindowIcon(QtGui.QIcon('player.png'))

        p = self.palette()
        p.setColor(QtGui.QPalette.Window, QtCore.Qt.black)
        self.setPalette(p)

        self.synchronize_viewer = None
        self.add_open_button = open_button
        self.play_position = NumericParameter()
        self.play_position.float_scale = 1000

        self.init_ui()
        # self.show()

    def init_ui(self):

        # create media player objectexamples_qtvlc.py
        self.mediaPlayer = QtMultimedia.QMediaPlayer(self)
        # self.mediaPlayer.setNotifyInterval(20)

        # create videowidget object

        # using graphics view may seem a good idea but the video is playing with lags
        use_graphic_view = False
        if use_graphic_view:
            video_item = QtMultimediaWidgets.QGraphicsVideoItem()
            self.video_view  = QtWidgets.QGraphicsView()
        else:
            use_scroll = False
            videowidget = myVideoWidget()
            if use_scroll:
                video_scroll = QtWidgets.QScrollArea()
                video_scroll.setWidget(videowidget)
            else:
                video_scroll = videowidget

        # create open button
        if self.add_open_button:
            openBtn = QtWidgets.QPushButton('Open Video')
            openBtn.clicked.connect(self.open_file)

        # create button for playing
        self.playBtn = QtWidgets.QPushButton()
        self.playBtn.setEnabled(False)
        self.playBtn.setIcon(self.style().standardIcon(QtWidgets.QStyle.SP_MediaPlay))
        self.playBtn.clicked.connect(lambda: self.synchronize_toggle_play(self))

        # create slider
        self.play_position_gui = NumericParameterGui(name="play_position", param=self.play_position,
                                                     callback=lambda: self.synchronize_set_play_position(self))
        self.play_position_gui.decimals = 3


        self.slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        self.slider.setRange(0, 0)
        self.slider.sliderMoved.connect(self.set_position)
        # create label
        self.label = QtWidgets.QLabel()
        self.label.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

        # create hbox layout
        hboxLayout = QtWidgets.QHBoxLayout()
        hboxLayout.setContentsMargins(0, 0, 0, 0)

        # set widgets to the hbox layout
        if self.add_open_button:
            hboxLayout.addWidget(openBtn)
        hboxLayout.addWidget(self.playBtn)
        hboxLayout.addWidget(self.slider)
        self.play_position_gui.add_to_layout(hboxLayout)

        # create vbox layout
        vboxLayout = QtWidgets.QVBoxLayout()

        if use_graphic_view:
            vboxLayout.addWidget(self.video_view)
            self.video_scene = QtWidgets.QGraphicsScene(0, 0, self.video_view.size().width(), self.video_view.size().height())
            self.video_view.setScene(self.video_scene)
            self.video_scene.addItem(video_item)
        else:
            vboxLayout.addWidget(video_scroll)

        vboxLayout.addLayout(hboxLayout)
        vboxLayout.addWidget(self.label)

        self.setLayout(vboxLayout)

        if use_graphic_view:
            self.mediaPlayer.setVideoOutput(video_item)
        else:
            self.mediaPlayer.setVideoOutput(videowidget)

        # media player signals

        self.mediaPlayer.playbackStateChanged.connect(self.mediastate_changed)
        self.mediaPlayer.positionChanged.connect(self.position_changed)
        self.mediaPlayer.durationChanged.connect(self.duration_changed)


    def set_synchronize(self, viewer):
        self.synchronize_viewer = viewer

    # def synchronize_data(self, other_viewer):
    #     other_viewer.current_scale = self.current_scale
    #     other_viewer.current_dx = self.current_dx
    #     other_viewer.current_dy = self.current_dy
    #     other_viewer.mouse_dx = self.mouse_dx
    #     other_viewer.mouse_dy = self.mouse_dy
    #     other_viewer.mouse_zx = self.mouse_zx
    #     other_viewer.mouse_zy = self.mouse_zy
    #     other_viewer.mouse_x = self.mouse_x
    #     other_viewer.mouse_y = self.mouse_y

    def synchronize_toggle_play(self, event_viewer):
        self.toggle_play_video()
        if self.synchronize_viewer is not None and self.synchronize_viewer is not event_viewer:
            self.synchronize_viewer.synchronize_toggle_play(event_viewer)

    def open_file(self):
        filename, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Open Video")

        if filename != '':
            self.mediaPlayer.setSource(QtCore.QUrl.fromLocalFile(filename))
            self.playBtn.setEnabled(True)

    def set_video(self, filename):
        if filename != '':
            self.mediaPlayer.setSource(QtCore.QUrl.fromLocalFile(filename))
            self.playBtn.setEnabled(True)

    def toggle_play_video(self):
        if self.mediaPlayer.playbackState() == QtMultimedia.QMediaPlayer.PlayingState:
            self.mediaPlayer.pause()
        else:
            self.mediaPlayer.play()

    def mediastate_changed(self, state):
        if self.mediaPlayer.playbackState() == QtMultimedia.QMediaPlayer.PlayingState:
            self.playBtn.setIcon( self.style().standardIcon(QtWidgets.QStyle.SP_MediaPause) )
            metadatalist = self.mediaPlayer.metaData()
            print(f"metadatalist = {metadatalist}")
            print(f"metadatalist = {metadatalist}")
            for key in metadatalist:
                var_data = self.mediaPlayer.metaData(key)
                print(f"{key}: {var_data}")
        else:
            self.playBtn.setIcon( self.style().standardIcon(QtWidgets.QStyle.SP_MediaPlay) )

    def position_changed(self, position):
        self.slider.setValue(position)
        self.play_position_gui.setValue(position)

    def duration_changed(self, duration):
        self.slider.setRange(0, duration)

        self.play_position.range = [0, duration]
        self.play_position_gui.setRange(0, duration)

    def set_position(self, position):
        self.mediaPlayer.setPosition(position)

    def synchronize_set_play_position(self, event_viewer):
        if self.mediaPlayer.playbackState() != QtMultimedia.QMediaPlayer.PlayingState:
            self.set_play_position()
            if self.synchronize_viewer is not None and self.synchronize_viewer is not event_viewer:
                self.synchronize_viewer.play_position.copy_from(self.play_position)
                self.synchronize_viewer.synchronize_set_play_position(event_viewer)

    def set_play_position(self):
        self.mediaPlayer.setPosition(self.play_position.int)

    def handle_errors(self):
        self.playBtn.setEnabled(False)
        self.label.setText("Error: " + self.mediaPlayer.errorString())


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    window = VideoPlayer(open_button=True)
    window.show()
    sys.exit(app.exec_())



from qimview.utils.viewer_image import *
import os
from .libraw_reader import libraw_supported_formats, read_libraw, has_rawpy
from .turbojpeg_reader import read_jpeg_turbojpeg, gb_turbo_jpeg, has_turbojpeg
from .simplejpeg_reader import read_jpeg_simplejpeg, has_simplejpeg
from .opencv_reader import read_opencv, opencv_supported_formats
from typing import Optional, TYPE_CHECKING

# Avoid circular imports
# Imports specific for type checking
if TYPE_CHECKING:
    from qimview.cache import FileCache

def read_jpeg(image_filename, image_buffer, read_size='full', use_RGB=True, verbose=False):
    # in terms of speed, it seems that turbojpeg is faster than simplejpeg which is faster than cv2
    # so turbojpeg > simplejpeg > cv2
    verbose = True
    im1 = im2 = im3 = None
    if gb_turbo_jpeg and has_turbojpeg:
        im1 = read_jpeg_turbojpeg(image_filename, image_buffer, read_size=read_size, use_RGB=use_RGB, verbose=verbose)
    if im1 is not None: return im1
    if read_size == 'full' and has_simplejpeg:
        im2 = read_jpeg_simplejpeg(image_filename, image_buffer, read_size, use_RGB, verbose)
    if im2 is not None: return im2
    im3 = read_opencv(image_filename, image_buffer, read_size, use_RGB, verbose)
    return im3


class ImageReader:
    def __init__(self):
        # set default plugins
        self._plugins = {
            ".JPG":read_jpeg,
            ".JPEG":read_jpeg,
        }
        # add libraw supported extensions
        if has_rawpy:
            for ext in libraw_supported_formats():
                if ext.upper() not in self._plugins:
                    self._plugins[ext.upper()] = read_libraw
        # add all opencv supposedly supported extensions
        for ext in opencv_supported_formats():
            if ext.upper() not in self._plugins:
                self._plugins[ext.upper()] = read_opencv
        self.file_cache : Optional[FileCache] = None

    def extensions(self):
        return list(self._plugins.keys())

    def set_file_cache(self, file_cache):
        self.file_cache = file_cache

    def set_plugin(self, extensions, callback):
        """ Set support to a image format based on list of extensions and callback

        Args:
            extensions ([type]): [description]
            callback (function): callback has signature (image_filename, image_buffer, read_size='full', use_RGB=True, verbose=False) 
            and returns a ViewerImage object
        """
        for ext in extensions:
            self._plugins[ext.upper()] = callback

    def read(self, filename, buffer=None, read_size='full', use_RGB=True, verbose=False, check_filecache_size=True):
        extension = os.path.splitext(filename)[1].upper()
        if extension not in self._plugins:
            print(  f"ERROR: ImageRead.read({filename}) extension not supported, "
                    f"supported extensions are {self._plugins.keys()}")
            return None

        try:
            if buffer is None and self.file_cache is not None:
                # try to get the buffer from the file cache
                buffer, fromcache = self.file_cache.get_file(filename, check_size=check_filecache_size)
                print(f" got buffer from cache? {fromcache}")
            res = self._plugins[extension](filename, buffer, read_size, use_RGB, verbose)
            if res:
                res.set_filename(filename)
            return res
        except Exception as e:
            print(f"Exception while reading image {filename}: {e}")
            return None

# unique instance of ImageReader for the application
gb_image_reader = ImageReader()


def reader_add_plugins():
    print("reader_add_plugins()")
    import configparser, sys
    config = configparser.ConfigParser()
    # config.read_file(open('default.cfg'))
    res = config.read([os.path.expanduser('~/.qimview.cfg')])
    # if read fails, found is probably not found
    if res:
        # Add new image format support
        try:
            formats = config['READERS']['Formats'].split(',')
        except Exception as e:
            print(f" ----- No reader plugin in config file: {e}")
        for fmt in  formats:
            try:
                format_cfg = config[f'READER.{fmt.upper()}']
                folder, module, ext = format_cfg['Folder'], format_cfg['Module'], format_cfg['Extensions'].split(',')
                folder = os.path.expanduser(folder)
                print(f' {fmt} {folder, ext}')
                # TODO: change this code, avoid sys.path.append()
                sys.path.append(folder)
                import importlib
                fmt_reader = importlib.import_module(f"{module}")
                gb_image_reader.set_plugin(ext, fmt_reader.read)
            except Exception as e:
                print(f" ----- Failed to add support for {fmt}: {e}")

reader_add_plugins()




from qimview.utils.utils import get_time
from qimview.image_readers import gb_image_reader
from .basecache import BaseCache
import os
import psutil
from typing import TYPE_CHECKING
from qimview.utils.viewer_image import ViewerImage

class ImageCache(BaseCache[str, ViewerImage, float]): 
    """
        Save output bytes from read() function into a cache indexed by the filename
        inherits from BaseCache, with
            id as string: input filename
            ViewerImage: image object
            mtime: modification time as float from osp.getmtime(filename)
        If a file is in the cache but its modification time on disk is more recent,
        we can enable an automatic reload

    Args:
        BaseCache (_type_): _description_
    """    
    def __init__(self):
        BaseCache.__init__(self, "ImageCache")
        # let use 25% of total memory
        total_memory = psutil.virtual_memory().total / self.cache_unit
        self.max_cache_size = int(total_memory * 0.25)
        self.verbose : bool = False

    def has_image(self, filename):
        # is it too slow
        filename = os.path.abspath(filename)
        return filename in self.cache_list

    def get_image(self, filename, read_size='full', verbose=False, use_RGB=True, image_transform=None,
                  check_size=True):
        """
        :param filename:
        :param show_timing:
        :return: pair image_data, boolean (True is coming from cache)
        """
        start = get_time()
        # Get absolute normalized path
        filename = os.path.abspath(filename)
        # print(f"image cache get_image({filename})")
        image_data = self.search(filename)
        mtime = os.path.getmtime(filename)

        if image_data is not None:
            if image_data[2] >= mtime:
                return image_data[1], True
            else:
                # Remove outdated cache element
                self.cache.remove(image_data)
                self.cache_list.remove(filename)
        
        image = gb_image_reader.read(filename, None, read_size, use_RGB=use_RGB, verbose=verbose,
                                            check_filecache_size=check_size)
        if image is not None:
            if image_transform is not None:
                image = image_transform(image)
            self.append(filename, image, extra=mtime, check_size=check_size)
            self._print_log(" get_image after read_image took {0:0.3f} sec.".format(get_time() - start))
        else:
            print(f"Failed to load image {filename}")
            return None, False
        return image, False

    def add_image(self, filename, read_size='full', verbose=False, use_RGB=True, image_transform=None,
                    progress_callback = None):
        """
        :param filename:
        :param show_timing:
        :return: pair image_data, boolean (True is coming from cache)
        """
        data, flag = self.get_image(filename, read_size, verbose, use_RGB, image_transform, check_size=False)
        return data is not None

    def image_added(self, filename):
        pass
        # print(f" *** Added image {os.path.basename(filename)} to cache finished")

    def add_result(self, r):
        # print(f"adding {r} to results ")
        self.add_results.append(r)

    def add_images(self, filenames, read_size='full', verbose=False, use_RGB=True, image_transform=None):
        start = get_time()
        self.add_results = []
        num_workers = 0
        for f in filenames:
            if f is not None and not self.has_image(f):
                # print(f" start worker with image {f}")
                self.thread_pool.set_worker(self.add_image, f, read_size, verbose, use_RGB, image_transform)
                self.thread_pool.set_worker_callbacks(finished_cb=lambda: self.image_added(f),
                                                      result_cb=self.add_result)
                self.thread_pool.start_worker()
                num_workers += 1
        wait_time = 0
        while len(self.add_results) != num_workers and wait_time < 2:
            # wait 5 ms before checking again
            #sleep(0.002)
            self.thread_pool.waitForDone(2)
            wait_time += 0.002
        self._print_log(f" ImageCache.add_images() wait_time {wait_time} took {int((get_time()-start)*1000+0.5)} ms;")
        if num_workers>0:
            self.thread_pool.clear()
            # There could be unfinished thread here?
            # It seems that the progress bar must be updated outside the threads
            self.check_size_limit(update_progress=True)
            if gb_image_reader.file_cache is not None:
                gb_image_reader.file_cache.check_size_limit(update_progress=True)
        assert len(self.add_results) == num_workers, "Workers left"
        self._print_log(f" ImageCache.add_images() {num_workers}, {self.add_results} took {int((get_time()-start)*1000+0.5)} ms;")



""" Simple TabDialog class """

from PySide6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QDialog,
    QTabWidget,
    QDialogButtonBox,
    QTextBrowser,
    QSizePolicy
)
from PySide6 import QtCore
from typing import Optional

class TabDialog(QDialog):
    def __init__(self, parent: Optional[QWidget] = None, title: str = "TabDialog") -> None:
        super().__init__(parent)
        # self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.tab_widget = QTabWidget()
        self.tab_widget.setUsesScrollButtons(False)
        # self.tab_widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        button_box = QDialogButtonBox( QDialogButtonBox.StandardButton.Ok )
        button_box.accepted.connect(self.accept)
        self.main_layout = QVBoxLayout()
        self.main_layout.addWidget(self.tab_widget)
        self.main_layout.addWidget(button_box)
        self.setLayout(self.main_layout)
        self.setWindowTitle(title)

    def add_tab(self, widget: QWidget, title:str)->None:
        self.tab_widget.addTab(widget, title)

    def add_markdown_tab(self, title:str, text: str)->None:
        text_browser = QTextBrowser()
        text_browser.setOpenExternalLinks(True)
        # # Avoid scroll bars for the moment
        # textedit.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        # text_browser.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        text_browser.setReadOnly(True)
        text_browser.setMarkdown(text)
        self.tab_widget.addTab(text_browser,title)
        # self.tab_widget.updateGeometry()

if __name__ == "__main__":
    from PySide6.QtWidgets import (
        QApplication,
        QTextEdit,
    )
    import sys
    # Usage example
    app = QApplication(sys.argv)
    tab_dialog = TabDialog()
    text1 = QTextEdit()
    text1.setReadOnly(True)
    text1.setMarkdown(" markdown text here  \n"
                      "second line")
    tab_dialog.add_tab(text1,"text1")

    text2 = QTextEdit()
    text2.setReadOnly(True)
    text2.setMarkdown("## section  \n"
                      " markdown text here  \n"
                      "second line")
    tab_dialog.add_tab(text2,"text2")
    tab_dialog.show()
    sys.exit(app.exec())
